#########################################
# Mixed Effect Model IPW
#########################################
from __future__ import division
import numpy as np
import pandas as pd
import statsmodels.api as sm

# Below is for SM GLIMMIX
from scipy.optimize import minimize
from scipy import sparse
import statsmodels.base.model as base
from statsmodels.iolib import summary2
from statsmodels.genmod import families
import warnings
import patsy


# TODO have it work
# TODO have different estimators (important for households of various sizes Basse & Feller 2018)
# TODO validate with R data
# TODO variance estimation
# TODO diagnostics


class InterferenceIPW:
    """IPW for interference settings. Described in Tchetgen Tchetgen and VanderWeele 2012, and Perez-Heydrich et al.
    2014.
    """
    print('InterferenceIPW is not supported yet')

    def __init__(self, stabilized=True, coverage_comparison=None, estimator='individual', alpha=0.05):
        """stuff"""
        self._estimator = estimator
        self._stabilized = stabilized
        self._coverage_ref = coverage_comparison
        self.alpha = alpha

    def fit(self, df, idvar, group, treatment, outcome, model, allocations):
        df = df.copy()
        if self._coverage_ref is None:
            reference = np.mean(df[treatment])
        else:
            reference = self._coverage_ref

        # Setting up allocations as needed
        comparison_allocation_j = self._pi_calc(alpha=reference, exclude_j=False)
        comparison_allocation_n = self._pi_calc(alpha=reference, exclude_j=True)

        # Getting predicted probabilities as needed
        mm = BinomialBayesMixedGLM.from_formula(outcome+' ~ '+model,  # TODO once 0.9.0 released, switch this
                                                vc_formulas={'cluster': group}, data=df).fit()
        # mm = sm.BinomialBayesMixedGLM.from_formula(outcome+' ~ art + male',
        #                                           vc_formulas={'label': 'age0'}, data=self.df).fit_map()
        df['pred'] = mm.predict()

        # estimating each of the corresponding functions
        for a in allocations:
            print(a)

    def _pi_calc(self, alpha, exclude_j=False):
        """
        Calculates the part of the numerator for Y^ipw_i as designated by pi(A; alpha) in Perez-Heydrich et al. 2014

        For estimating Y(alpha)

            pi(A_i ; alpha) = cumprod_j=1(alpha^A * (1-alpha)^(1-A))

        For estimating Y(a;alpha) need to exclude j from the calculation such that;

            pi(A_i,-j ; alpha) = cumprod_k=1,k!=j(alpha^A * (1-alpha)^(1-A))

        This is done by using that same process as before, but now we divide by the individual level (to cancel out that
        cumulative product including individual j
        """
        pf = self.df.copy()
        pf['i_alpha'] = alpha ** (pf['treatment']) * (1 - alpha) ** (1 - pf['treatment'])
        pf.groupby(self.group)['i_alpha'].prod().reset_index()
        pf = pd.merge(pf, pf.groupby(self.group)['i_alpha'].prod().rename('g_alpha').reset_index(),
                      left_on=self.group, right_on=self.group)

        if exclude_j:
            pi = pf['g_alpha'] / pf['i_alpha']
        else:
            pi = pf['g_alpha']
        return pi

    def _yipw_calc(self, denom, a=None):
        """
        This calculates the corresponding Y^ipw for each potential combination. The version used depends on if 'a' is
        specified

        Y^ipw_i(a; alpha) = (sum(pi(A_i,-j; alpha) * I(A=a) * Y^ij) / Pr(A_i|X_i; phi) * n_i

        Y^ipw_i(alpha) = (sum(pi(A_i; alpha) * Y^ij) / Pr(A_i|X_i; phi) * n_i
        """
        ni = self.df.groupby(self.group)[self.id].count()
        if a is None:
            self.df['raw_numerator'] = self.df['pi2'] * self.df[self.ou]
            numer = self.df.groupby(self.group)['raw_numerator'].sum()
            yipw = numer / (denom * ni)
        else:
            self.df['raw_numerator'] = self.df['pi1'] * np.where(self.df[self.ex]==treat, 1, 0) * self.df[self.ou]
            numer = self.df.groupby(self.group)['raw_numerator'].sum()
            yipw = numer / (denom * ni)
        return yipw


#################################################
# Statsmodels GLIMMIX purely for my testing
# here until new statsmodels version released
#################################################
glw = [
    [0.2955242247147529, -0.1488743389816312],
    [0.2955242247147529, 0.1488743389816312],
    [0.2692667193099963, -0.4333953941292472],
    [0.2692667193099963, 0.4333953941292472],
    [0.2190863625159820, -0.6794095682990244],
    [0.2190863625159820, 0.6794095682990244],
    [0.1494513491505806, -0.8650633666889845],
    [0.1494513491505806, 0.8650633666889845],
    [0.0666713443086881, -0.9739065285171717],
    [0.0666713443086881, 0.9739065285171717],
]


class _BayesMixedGLM(base.Model):
    def __init__(self,
                 endog,
                 exog,
                 exog_vc=None,
                 ident=None,
                 family=None,
                 vcp_p=1,
                 fe_p=2,
                 fep_names=None,
                 vcp_names=None,
                 vc_names=None,
                 **kwargs):

        if len(ident) != exog_vc.shape[1]:
            msg = "len(ident) should match the number of columns of exog_vc"
            raise ValueError(msg)

        # Get the fixed effects parameter names
        if fep_names is None:
            if hasattr(exog, "columns"):
                fep_names = exog.columns.tolist()
            else:
                fep_names = ["FE_%d" % (k + 1) for k in range(exog.shape[1])]

        # Get the variance parameter names
        if vcp_names is None:
            vcp_names = ["VC_%d" % (k + 1) for k in range(int(max(ident)) + 1)]
        else:
            if len(vcp_names) != len(set(ident)):
                msg = "The lengths of vcp_names and ident should be the same"
                raise ValueError(msg)

        if not sparse.issparse(exog_vc):
            exog_vc = sparse.csr_matrix(exog_vc)

        ident = ident.astype(np.int)
        vcp_p = float(vcp_p)
        fe_p = float(fe_p)

        # Number of fixed effects parameters
        if exog is None:
            k_fep = 0
        else:
            k_fep = exog.shape[1]

        # Number of variance component structure parameters and
        # variance component realizations.
        if exog_vc is None:
            k_vc = 0
            k_vcp = 0
        else:
            k_vc = exog_vc.shape[1]
            k_vcp = max(ident) + 1

        # power would be better but not available in older scipy
        exog_vc2 = exog_vc.multiply(exog_vc)

        super(_BayesMixedGLM, self).__init__(endog, exog, **kwargs)

        self.exog_vc = exog_vc
        self.exog_vc2 = exog_vc2
        self.ident = ident
        self.family = family
        self.k_fep = k_fep
        self.k_vc = k_vc
        self.k_vcp = k_vcp
        self.fep_names = fep_names
        self.vcp_names = vcp_names
        self.vc_names = vc_names
        self.fe_p = fe_p
        self.vcp_p = vcp_p
        self.names = fep_names + vcp_names
        if vc_names is not None:
            self.names += vc_names

    def _unpack(self, vec):

        ii = 0

        # Fixed effects parameters
        fep = vec[:ii + self.k_fep]
        ii += self.k_fep

        # Variance component structure parameters (standard
        # deviations).  These are on the log scale.  The standard
        # deviation for random effect j is exp(vcp[ident[j]]).
        vcp = vec[ii:ii + self.k_vcp]
        ii += self.k_vcp

        # Random effect realizations
        vc = vec[ii:]

        return fep, vcp, vc

    def logposterior(self, params):
        """
        The overall log-density: log p(y, fe, vc, vcp).
        This differs by an additive constant from the log posterior
        log p(fe, vc, vcp | y).
        """

        fep, vcp, vc = self._unpack(params)

        # Contributions from p(y | x, vc)
        lp = 0
        if self.k_fep > 0:
            lp += np.dot(self.exog, fep)
        if self.k_vc > 0:
            lp += self.exog_vc.dot(vc)

        mu = self.family.link.inverse(lp)
        ll = self.family.loglike(self.endog, mu)

        if self.k_vc > 0:

            # Contributions from p(vc | vcp)
            vcp0 = vcp[self.ident]
            s = np.exp(vcp0)
            ll -= 0.5 * np.sum(vc**2 / s**2) + np.sum(vcp0)

            # Contributions from p(vc)
            ll -= 0.5 * np.sum(vcp**2 / self.vcp_p**2)

        # Contributions from p(fep)
        if self.k_fep > 0:
            ll -= 0.5 * np.sum(fep**2 / self.fe_p**2)

        return ll

    def logposterior_grad(self, params):
        """
        The gradient of the log posterior.
        """

        fep, vcp, vc = self._unpack(params)

        lp = 0
        if self.k_fep > 0:
            lp += np.dot(self.exog, fep)
        if self.k_vc > 0:
            lp += self.exog_vc.dot(vc)

        mu = self.family.link.inverse(lp)

        score_factor = (self.endog - mu) / self.family.link.deriv(mu)
        score_factor /= self.family.variance(mu)

        te = [None, None, None]

        # Contributions from p(y | x, z, vc)
        if self.k_fep > 0:
            te[0] = np.dot(score_factor, self.exog)
        if self.k_vc > 0:
            te[2] = self.exog_vc.transpose().dot(score_factor)

        if self.k_vc > 0:
            # Contributions from p(vc | vcp)
            # vcp0 = vcp[self.ident]
            # s = np.exp(vcp0)
            # ll -= 0.5 * np.sum(vc**2 / s**2) + np.sum(vcp0)
            vcp0 = vcp[self.ident]
            s = np.exp(vcp0)
            u = vc**2 / s**2 - 1
            te[1] = np.bincount(self.ident, weights=u)
            te[2] -= vc / s**2

            # Contributions from p(vcp)
            # ll -= 0.5 * np.sum(vcp**2 / self.vcp_p**2)
            te[1] -= vcp / self.vcp_p**2

        # Contributions from p(fep)
        if self.k_fep > 0:
            te[0] -= fep / self.fe_p**2

        te = [x for x in te if x is not None]

        return np.concatenate(te)

    def _get_start(self):
        start_fep = np.zeros(self.k_fep)
        start_vcp = np.ones(self.k_vcp)
        start_vc = np.random.normal(size=self.k_vc)
        start = np.concatenate((start_fep, start_vcp, start_vc))
        return start

    @classmethod
    def from_formula(cls,
                     formula,
                     vc_formulas,
                     data,
                     family=None,
                     vcp_p=1,
                     fe_p=2):
        """
        Fit a BayesMixedGLM using a formula.
        Parameters
        ----------
        formula : string
            Formula for the endog and fixed effects terms (use ~ to separate
            dependent and independent expressions).
        vc_formulas : dictionary
            vc_formulas[name] is a one-sided formula that creates one
            collection of random effects with a common variance
            prameter.  If using a categorical expression to produce
            variance components, note that generally `0 + ...` should
            be used so that an intercept is not included.
        data : data frame
            The data to which the formulas are applied.
        family : genmod.families instance
            A GLM family.
        vcp_p : float
            The prior standard deviation for the logarithms of the standard
            deviations of the random effects.
        fe_p : float
            The prior standard deviation for the fixed effects parameters.
        """

        ident = []
        exog_vc = []
        vcp_names = []
        j = 0
        for na, fml in vc_formulas.items():
            mat = patsy.dmatrix(fml, data, return_type='dataframe')
            exog_vc.append(mat)
            vcp_names.append(na)
            ident.append(j * np.ones(mat.shape[1]))
            j += 1
        exog_vc = pd.concat(exog_vc, axis=1)
        vc_names = exog_vc.columns.tolist()

        ident = np.concatenate(ident)

        model = super(_BayesMixedGLM, cls).from_formula(
            formula,
            data=data,
            family=family,
            subset=None,
            exog_vc=exog_vc,
            ident=ident,
            vc_names=vc_names,
            vcp_names=vcp_names,
            fe_p=fe_p,
            vcp_p=vcp_p)

        return model

    def fit(self, method="BFGS", minim_opts=None):
        """
        fit is equivalent to fit_map for this class.
        See fit_map for parameter information.
        """
        self.fit_map(method, minim_opts)

    def fit_map(self, method="BFGS", minim_opts=None, scale_fe=False):
        """
        Construct the Laplace approximation to the posterior
        distribution.
        Parameters
        ----------
        method : string
            Optimization method for finding the posterior mode.
        minim_opts : dict-like
            Options passed to scipy.minimize.
        scale_fe : bool
            If true, the columns of the fixed effects design matrix
            are centered and scaled to unit variance before fitting
            the model.  The results are back-transformed so that the
            results are presented on the original scale.
        Returns
        -------
        BayesMixedGLMResults instance.
        """

        if scale_fe:
            mn = self.exog.mean(0)
            sc = self.exog.std(0)
            self._exog_save = self.exog
            self.exog = self.exog.copy()
            ixs = np.flatnonzero(sc > 1e-8)
            self.exog[:, ixs] -= mn[ixs]
            self.exog[:, ixs] /= sc[ixs]

        def fun(params):
            return -self.logposterior(params)

        def grad(params):
            return -self.logposterior_grad(params)

        start = self._get_start()

        r = minimize(fun, start, method=method, jac=grad, options=minim_opts)
        if not r.success:
            msg = ("Laplace fitting did not converge, |gradient|=%.6f" %
                   np.sqrt(np.sum(r.jac**2)))
            warnings.warn(msg)

        from statsmodels.tools.numdiff import approx_fprime
        hess = approx_fprime(r.x, grad)
        cov = np.linalg.inv(hess)

        params = r.x

        if scale_fe:
            self.exog = self._exog_save
            del self._exog_save
            params[ixs] /= sc[ixs]
            cov[ixs, :][:, ixs] /= np.outer(sc[ixs], sc[ixs])

        return BayesMixedGLMResults(self, params, cov, optim_retvals=r)

    def predict(self, params, exog=None, linear=False):
        """
        Return the fitted mean structure.
        Parameters
        ----------
        params : array-like
            The parameter vector, may be the full parameter vector, or may
            be truncated to include only the mean parameters.
        exog : array-like
            The design matrix for the mean structure.  If omitted, use the
            model's design matrix.
        linear : bool
            If True, return the linear predictor without passing through the
            link function.
        Result
        ------
        A 1-dimensional array of predicted values
        """

        if exog is None:
            exog = self.exog

        q = exog.shape[1]
        pr = np.dot(exog, params[0:q])

        if not linear:
            pr = self.family.link.inverse(pr)

        return pr


class _VariationalBayesMixedGLM(object):
    """
    A mixin providing generic (not family-specific) methods for
    variational Bayes mean field fitting.
    """

    # Integration range (from -rng to +rng).  The integrals are with
    # respect to a standard Gaussian distribution so (-5, 5) will be
    # sufficient in many cases.
    rng = 5

    verbose = False

    # Returns the mean and variance of the linear predictor under the
    # given distribution parameters.
    def _lp_stats(self, fep_mean, fep_sd, vc_mean, vc_sd):

        tm = np.dot(self.exog, fep_mean)
        tv = np.dot(self.exog**2, fep_sd**2)
        tm += self.exog_vc.dot(vc_mean)
        tv += self.exog_vc2.dot(vc_sd**2)

        return tm, tv

    def vb_elbo_base(self, h, tm, fep_mean, vcp_mean, vc_mean, fep_sd, vcp_sd,
                     vc_sd):
        """
        Returns the evidence lower bound (ELBO) for the model.
        This function calculates the family-specific ELBO function
        based on information provided from a subclass.
        Parameters
        ----------
        h : function mapping 1d vector to 1d vector
            The contribution of the model to the ELBO function can be
            expressed as y_i*lp_i + Eh_i(z), where y_i and lp_i are
            the response and linear predictor for observation i, and z
            is a standard normal rangom variable.  This formulation
            can be achieved for any GLM with a canonical link
            function.
        """

        # p(y | vc) contributions
        iv = 0
        for w in glw:
            z = self.rng * w[1]
            iv += w[0] * h(z) * np.exp(-z**2 / 2)
        iv /= np.sqrt(2 * np.pi)
        iv *= self.rng
        iv += self.endog * tm
        iv = iv.sum()

        # p(vc | vcp) * p(vcp) * p(fep) contributions
        iv += self._elbo_common(fep_mean, fep_sd, vcp_mean, vcp_sd, vc_mean,
                                vc_sd)

        r = (iv + np.sum(np.log(fep_sd)) + np.sum(np.log(vcp_sd)) + np.sum(
            np.log(vc_sd)))

        return r

    def vb_elbo_grad_base(self, h, tm, tv, fep_mean, vcp_mean, vc_mean, fep_sd,
                          vcp_sd, vc_sd):
        """
        Return the gradient of the ELBO function.
        See vb_elbo_base for parameters.
        """

        fep_mean_grad = 0.
        fep_sd_grad = 0.
        vcp_mean_grad = 0.
        vcp_sd_grad = 0.
        vc_mean_grad = 0.
        vc_sd_grad = 0.

        # p(y | vc) contributions
        for w in glw:
            z = self.rng * w[1]
            u = h(z) * np.exp(-z**2 / 2) / np.sqrt(2 * np.pi)
            r = u / np.sqrt(tv)
            fep_mean_grad += w[0] * np.dot(u, self.exog)
            vc_mean_grad += w[0] * self.exog_vc.transpose().dot(u)
            fep_sd_grad += w[0] * z * np.dot(r, self.exog**2 * fep_sd)
            v = self.exog_vc2.multiply(vc_sd).transpose().dot(r)
            v = np.squeeze(np.asarray(v))
            vc_sd_grad += w[0] * z * v

        fep_mean_grad *= self.rng
        vc_mean_grad *= self.rng
        fep_sd_grad *= self.rng
        vc_sd_grad *= self.rng
        fep_mean_grad += np.dot(self.endog, self.exog)
        vc_mean_grad += self.exog_vc.transpose().dot(self.endog)

        (fep_mean_grad_i, fep_sd_grad_i, vcp_mean_grad_i, vcp_sd_grad_i,
         vc_mean_grad_i, vc_sd_grad_i) = self._elbo_grad_common(
             fep_mean, fep_sd, vcp_mean, vcp_sd, vc_mean, vc_sd)

        fep_mean_grad += fep_mean_grad_i
        fep_sd_grad += fep_sd_grad_i
        vcp_mean_grad += vcp_mean_grad_i
        vcp_sd_grad += vcp_sd_grad_i
        vc_mean_grad += vc_mean_grad_i
        vc_sd_grad += vc_sd_grad_i

        fep_sd_grad += 1 / fep_sd
        vcp_sd_grad += 1 / vcp_sd
        vc_sd_grad += 1 / vc_sd

        mean_grad = np.concatenate((fep_mean_grad, vcp_mean_grad,
                                    vc_mean_grad))
        sd_grad = np.concatenate((fep_sd_grad, vcp_sd_grad, vc_sd_grad))

        if self.verbose:
            print(
                "|G|=%f" % np.sqrt(np.sum(mean_grad**2) + np.sum(sd_grad**2)))

        return mean_grad, sd_grad

    def fit_vb(self,
               mean=None,
               sd=None,
               fit_method="BFGS",
               minim_opts=None,
               scale_fe=False,
               verbose=False):
        """
        Fit a model using the variational Bayes mean field approximation.
        Parameters
        ----------
        mean : array-like
            Starting value for VB mean vector
        sd : array-like
            Starting value for VB standard deviation vector
        fit_method : string
            Algorithm for scipy.minimize
        minim_opts : dict-like
            Options passed to scipy.minimize
        scale_fe : bool
            If true, the columns of the fixed effects design matrix
            are centered and scaled to unit variance before fitting
            the model.  The results are back-transformed so that the
            results are presented on the original scale.
        verbose : bool
            If True, print the gradient norm to the screen each time
            it is calculated.
        Notes
        -----
        The goal is to find a factored Gaussian approximation
        q1*q2*...  to the posterior distribution, approximately
        minimizing the KL divergence from the factored approximation
        to the actual posterior.  The KL divergence, or ELBO function
        has the form
            E* log p(y, fe, vcp, vc) - E* log q
        where E* is expectation with respect to the product of qj.
        References
        ----------
        Blei, Kucukelbir, McAuliffe (2017).  Variational Inference: A
        review for Statisticians
        https://arxiv.org/pdf/1601.00670.pdf
        """

        self.verbose = verbose

        if scale_fe:
            mn = self.exog.mean(0)
            sc = self.exog.std(0)
            self._exog_save = self.exog
            self.exog = self.exog.copy()
            ixs = np.flatnonzero(sc > 1e-8)
            self.exog[:, ixs] -= mn[ixs]
            self.exog[:, ixs] /= sc[ixs]

        n = self.k_fep + self.k_vcp + self.k_vc
        ml = self.k_fep + self.k_vcp + self.k_vc
        if mean is None:
            m = np.zeros(n)
        else:
            if len(mean) != ml:
                raise ValueError(
                    "mean has incorrect length, %d != %d" % (len(mean), ml))
            m = mean.copy()
        if sd is None:
            s = -0.5 + 0.1 * np.random.normal(size=n)
        else:
            if len(sd) != ml:
                raise ValueError(
                    "sd has incorrect length, %d != %d" % (len(sd), ml))

            # s is parameterized on the log-scale internally when
            # optimizing the ELBO function (this is transparent to the
            # caller)
            s = np.log(sd)

        # Don't allow the variance parameter starting mean values to
        # be too small.
        i1, i2 = self.k_fep, self.k_fep + self.k_vcp
        m[i1:i2] = np.where(m[i1:i2] < -1, -1, m[i1:i2])

        # Don't allow the posterior standard deviation starting values
        # to be too small.
        s = np.where(s < -1, -1, s)

        def elbo(x):
            n = len(x) // 2
            return -self.vb_elbo(x[:n], np.exp(x[n:]))

        def elbo_grad(x):
            n = len(x) // 2
            gm, gs = self.vb_elbo_grad(x[:n], np.exp(x[n:]))
            gs *= np.exp(x[n:])
            return -np.concatenate((gm, gs))

        start = np.concatenate((m, s))
        mm = minimize(
            elbo, start, jac=elbo_grad, method=fit_method, options=minim_opts)
        if not mm.success:
            warnings.warn("VB fitting did not converge")

        n = len(mm.x) // 2
        params = mm.x[0:n]
        va = np.exp(2 * mm.x[n:])

        if scale_fe:
            self.exog = self._exog_save
            del self._exog_save
            params[ixs] /= sc[ixs]
            va[ixs] /= sc[ixs]**2

        return BayesMixedGLMResults(self, params, va, mm)

    # Handle terms in the ELBO that are common to all models.
    def _elbo_common(self, fep_mean, fep_sd, vcp_mean, vcp_sd, vc_mean, vc_sd):

        iv = 0

        # p(vc | vcp) contributions
        m = vcp_mean[self.ident]
        s = vcp_sd[self.ident]
        iv -= np.sum((vc_mean**2 + vc_sd**2) * np.exp(2 * (s**2 - m))) / 2
        iv -= np.sum(m)

        # p(vcp) contributions
        iv -= 0.5 * (vcp_mean**2 + vcp_sd**2).sum() / self.vcp_p**2

        # p(b) contributions
        iv -= 0.5 * (fep_mean**2 + fep_sd**2).sum() / self.fe_p**2

        return iv

    def _elbo_grad_common(self, fep_mean, fep_sd, vcp_mean, vcp_sd, vc_mean,
                          vc_sd):

        # p(vc | vcp) contributions
        m = vcp_mean[self.ident]
        s = vcp_sd[self.ident]
        u = vc_mean**2 + vc_sd**2
        ve = np.exp(2 * (s**2 - m))
        dm = u * ve - 1
        ds = -2 * u * ve * s
        vcp_mean_grad = np.bincount(self.ident, weights=dm)
        vcp_sd_grad = np.bincount(self.ident, weights=ds)

        vc_mean_grad = -vc_mean.copy() * ve
        vc_sd_grad = -vc_sd.copy() * ve

        # p(vcp) contributions
        vcp_mean_grad -= vcp_mean / self.vcp_p**2
        vcp_sd_grad -= vcp_sd / self.vcp_p**2

        # p(b) contributions
        fep_mean_grad = -fep_mean.copy() / self.fe_p**2
        fep_sd_grad = -fep_sd.copy() / self.fe_p**2

        return (fep_mean_grad, fep_sd_grad, vcp_mean_grad, vcp_sd_grad,
                vc_mean_grad, vc_sd_grad)


class BayesMixedGLMResults(object):
    """
    Attributes
    ----------
    fe_mean : array-like
        Posterior mean of the fixed effects coefficients.
    fe_sd : array-like
        Posterior standard deviation of the fixed effects coefficients
    vcp_mean : array-like
        Posterior mean of the logged variance component standard
        deviations.
    vcp_sd : array-like
        Posterior standard deviation of the logged variance component
        standard deviations.
    vc_mean : array-like
        Posterior mean of the random coefficients
    vc_sd : array-like
        Posterior standard deviation of the random coefficients
    """

    def __init__(self, model, params, cov_params, optim_retvals=None):

        self.model = model
        self.params = params
        self._cov_params = cov_params
        self.optim_retvals = optim_retvals

        self.fe_mean, self.vcp_mean, self.vc_mean = (model._unpack(params))

        if cov_params.ndim == 2:
            cp = np.diag(cov_params)
        else:
            cp = cov_params
        self.fe_sd, self.vcp_sd, self.vc_sd = model._unpack(cp)
        self.fe_sd = np.sqrt(self.fe_sd)
        self.vcp_sd = np.sqrt(self.vcp_sd)
        self.vc_sd = np.sqrt(self.vc_sd)

    def cov_params(self):

        if hasattr(self.model.data, "frame"):
            # Return the covariance matrix as a dataframe or series
            na = (self.model.fep_names + self.model.vcp_names +
                  self.model.vc_names)
            if self._cov_params.ndim == 2:
                return pd.DataFrame(self._cov_params, index=na, columns=na)
            else:
                return pd.Series(self._cov_params, index=na)

        # Return the covariance matrix as a ndarray
        return self._cov_params

    def summary(self):

        df = pd.DataFrame()
        m = self.model.k_fep + self.model.k_vcp
        df["Type"] = (["M" for k in range(self.model.k_fep)] +
                      ["V" for k in range(self.model.k_vcp)])

        df["Post. Mean"] = self.params[0:m]

        if self._cov_params.ndim == 2:
            v = np.diag(self._cov_params)[0:m]
            df["Post. SD"] = np.sqrt(v)
        else:
            df["Post. SD"] = np.sqrt(self._cov_params[0:m])

        # Convert variance parameters to natural scale
        df["SD"] = np.exp(df["Post. Mean"])
        df["SD (LB)"] = np.exp(df["Post. Mean"] - 2 * df["Post. SD"])
        df["SD (UB)"] = np.exp(df["Post. Mean"] + 2 * df["Post. SD"])
        df["SD"] = ["%.3f" % x for x in df.SD]
        df["SD (LB)"] = ["%.3f" % x for x in df["SD (LB)"]]
        df["SD (UB)"] = ["%.3f" % x for x in df["SD (UB)"]]
        df.loc[df.index < self.model.k_fep, "SD"] = ""
        df.loc[df.index < self.model.k_fep, "SD (LB)"] = ""
        df.loc[df.index < self.model.k_fep, "SD (UB)"] = ""

        df.index = self.model.fep_names + self.model.vcp_names

        summ = summary2.Summary()
        summ.add_title(self.model.family.__class__.__name__ +
                       " Mixed GLM Results")
        summ.add_df(df)

        summ.add_text("Parameter types are mean structure (M) and variance structure (V)")
        summ.add_text("Variance parameters are modeled as log standard deviations")

        return summ

    def random_effects(self, term=None):
        """
        Posterior mean and standard deviation of random effects.
        Parameters
        ----------
        term : int or None
            If None, results for all random effects are returned.  If
            an integer, returns results for a given set of random
            effects.  The value of `term` refers to an element of the
            `ident` vector, or to a position in the `vc_formulas`
            list.
        Returns
        -------
        Data frame of posterior means and posterior standard
        deviations of random effects.
        """

        z = self.vc_mean
        s = self.vc_sd
        na = self.model.vc_names

        if term is not None:
            termix = self.model.vcp_names.index(term)
            ii = np.flatnonzero(self.model.ident == termix)
            z = z[ii]
            s = s[ii]
            na = [na[i] for i in ii]

        x = pd.DataFrame({"Mean": z, "SD": s})

        if na is not None:
            x.index = na

        return x

    def predict(self, exog=None, linear=False):
        """
        Return predicted values for the mean structure.
        Parameters
        ----------
        exog : array-like
            The design matrix for the mean structure.  If None,
            use the model's design matrix.
        linear : bool
            If True, returns the linear predictor, otherwise
            transform the linear predictor using the link function.
        Returns
        -------
        A one-dimensional array of fitted values.
        """

        return self.model.predict(self.params, exog, linear)


class BinomialBayesMixedGLM(_VariationalBayesMixedGLM, _BayesMixedGLM):
    def __init__(self,
                 endog,
                 exog,
                 exog_vc,
                 ident,
                 vcp_p=1,
                 fe_p=2,
                 fep_names=None,
                 vcp_names=None,
                 vc_names=None):

        super(BinomialBayesMixedGLM, self).__init__(
            endog,
            exog,
            exog_vc=exog_vc,
            ident=ident,
            vcp_p=vcp_p,
            fe_p=fe_p,
            family=families.Binomial(),
            fep_names=fep_names,
            vcp_names=vcp_names,
            vc_names=vc_names)

    @classmethod
    def from_formula(cls, formula, vc_formulas, data, vcp_p=1, fe_p=2):

        fam = families.Binomial()
        x = _BayesMixedGLM.from_formula(
            formula, vc_formulas, data, family=fam, vcp_p=vcp_p, fe_p=fe_p)

        # Copy over to the intended class structure
        mod = BinomialBayesMixedGLM(
            x.endog,
            x.exog,
            exog_vc=x.exog_vc,
            ident=x.ident,
            vcp_p=x.vcp_p,
            fe_p=x.fe_p,
            fep_names=x.fep_names,
            vcp_names=x.vcp_names,
            vc_names=x.vc_names)
        mod.data = x.data

        return mod

    def vb_elbo(self, vb_mean, vb_sd):
        """
        Returns the evidence lower bound (ELBO) for the model.
        """

        fep_mean, vcp_mean, vc_mean = self._unpack(vb_mean)
        fep_sd, vcp_sd, vc_sd = self._unpack(vb_sd)
        tm, tv = self._lp_stats(fep_mean, fep_sd, vc_mean, vc_sd)

        def h(z):
            return -np.log(1 + np.exp(tm + np.sqrt(tv) * z))

        return self.vb_elbo_base(h, tm, fep_mean, vcp_mean, vc_mean, fep_sd,
                                 vcp_sd, vc_sd)

    def vb_elbo_grad(self, vb_mean, vb_sd):
        """
        Returns the gradient of the model's evidence lower bound (ELBO).
        """

        fep_mean, vcp_mean, vc_mean = self._unpack(vb_mean)
        fep_sd, vcp_sd, vc_sd = self._unpack(vb_sd)
        tm, tv = self._lp_stats(fep_mean, fep_sd, vc_mean, vc_sd)

        def h(z):
            u = tm + np.sqrt(tv) * z
            x = np.zeros_like(u)
            ii = np.flatnonzero(u > 0)
            uu = u[ii]
            x[ii] = 1 / (1 + np.exp(-uu))
            ii = np.flatnonzero(u <= 0)
            uu = u[ii]
            x[ii] = np.exp(uu) / (1 + np.exp(uu))
            return -x

        return self.vb_elbo_grad_base(h, tm, tv, fep_mean, vcp_mean, vc_mean,
                                      fep_sd, vcp_sd, vc_sd)



import warnings
import numpy as np
import statsmodels.api as sm
import statsmodels.formula.api as smf
from zepid.causal.ipw import propensity_score


class AIPTW:
    def __init__(self, df, exposure, outcome):
        """Implementation of augmented inverse probablity weight estimator. This is a simple implementation of AIPW
        and does not account for missing data. It only handles time-fixed confounders and treatments at this point.

        The augment-inverse probability weights are calculated from the following formula

        .. math::

            RD = (\frac{Y_{A=1}*A}{\Pr(A=1)} - \frac{\hat{Y}_1*(A-\Pr(A=1)}{\Pr(A=1)}) - (\frac{Y_{A=0}*A}{\Pr(A=0)
            } - \frac{\hat{Y}_0*(A-\Pr(A=1)}{\Pr(A=0)})
            RR = (\frac{Y_{A=1}*A}{\Pr(A=1)} - \frac{\hat{Y}_1*(A-\Pr(A=1)}{\Pr(A=1)}) / (\frac{Y_{A=0}*A}{\Pr(A=0)
            } - \frac{\hat{Y}_0*(A-\Pr(A=1)}{\Pr(A=0)})

        Confidence intervals should be calculated using non-parametric bootstrapping

        Parameters
        ----------
        df : DataFrame
            Pandas DataFrame object containing all variables of interest
        exposure : str
            Column name of the exposure variable. Currently only binary is supported
        outcome : str
            Column name of the outcome variable. Currently only binary is supported

        Examples
        --------
        Set up the environment and the data set
        >>>from zepid import load_sample_data, spline
        >>>from zepid.causal.doublyrobust import AIPTW
        >>>df = load_sample_data(timevary=False)
        >>>df[['cd4_rs1','cd4_rs2']] = spline(df,'cd40',n_knots=3,term=2,restricted=True)
        >>>df[['age_rs1','age_rs2']] = spline(df,'age0',n_knots=3,term=2,restricted=True)

        Initialize the AIPTW model
        >>>aipw = AIPTW(df, exposure='art', outcome='dead')

        Specify the exposure/treatment model
        >>>aipw.exposure_model('male + age0 + age_rs1 + age_rs2 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')

        Specify the outcome model
        >>>aipw.outcome_model('art + male + age0 + age_rs1 + age_rs2 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')

        Estimate the risk difference and risk ratio
        >>>aipw.fit()

        Displaying the results
        >>>aipw.summary()

        Extracting risk difference and risk ratio respectively
        >>>aipw.risk_difference
        >>>aipw.risk_ratio

        References
        ----------
        Funk, M. J., Westreich, D., Wiesen, C., Stürmer, T., Brookhart, M. A., & Davidian, M. (2011). Doubly robust
        estimation of causal effects. American Journal of Epidemiology, 173(7), 761-767.
        """
        self.df = df.copy()
        if df.dropna().shape[0] != df.shape[0]:
            warnings.warn("There is missing data in the dataset. By default, AIPW will drop all missing data. AIPW will"
                          "fit "+str(df.dropna().shape[0])+' of '+str(df.shape[0])+' observations', UserWarning)
        self.df = df.copy().dropna().reset_index()
        self._exposure = exposure
        self._outcome = outcome
        self._fit_exposure_model = False
        self._fit_outcome_model = False
        self._generated_ci = False
        self.risk_difference = None
        self.risk_ratio = None
        self._exp_model = None
        self._out_model = None

    def exposure_model(self, model, print_results=True):
        """Specify the propensity score / inverse probability weight model. Model used to predict the exposure via a
        logistic regression model

        Parameters
        ----------
        model : str
            Independent variables to predict the exposure. For example, 'var1 + var2 + var3'
        print_results : bool, optional
            Whether to print the fitted model results. Default is True (prints results)
        """
        self._exp_model = self._exposure + ' ~ ' + model
        fitmodel = propensity_score(self.df, self._exp_model, print_results=print_results)
        self.df['ps'] = fitmodel.predict(self.df)
        self._fit_exposure_model = True

    def outcome_model(self, model, print_results=True):
        """Specify the outcome model. Model used to predict the outcome via a logistic regression model

        Parameters
        ----------
        model : str
            Independent variables to predict the outcome. For example, 'var1 + var2 + var3 + var4'
        print_results : bool, optional
            Whether to print the fitted model results. Default is True (prints results)
        """
        self._out_model = self._outcome + ' ~ ' + model
        f = sm.families.family.Binomial()
        log = smf.glm(self._out_model, self.df, family=f).fit()
        if print_results:
            print('\n----------------------------------------------------------------')
            print('MODEL: ' + self._out_model)
            print('-----------------------------------------------------------------')
            print(log.summary())

        dfx = self.df.copy()
        dfx[self._exposure] = 1
        self.df['pY1'] = log.predict(dfx)
        dfx = self.df.copy()
        dfx[self._exposure] = 0
        self.df['pY0'] = log.predict(dfx)
        self._fit_outcome_model = True

    def fit(self):
        """Once the exposure and outcome models are specified, we can estimate the Risk Ratio and Risk Difference.
        This function generates the estimated risk difference and risk ratio. For confidence intervals, a non-parametric
        bootstrap procedure should be used
        """
        if (self._fit_exposure_model is False) or (self._fit_outcome_model is False):
            raise ValueError('The exposure and outcome models must be specified before the doubly robust estimate can '
                             'be generated')

        # Doubly robust estimator under all treated
        self.df['dr1'] = np.where(self.df[self._exposure] == 1,
                                  ((self.df[self._outcome]) / self.df['ps']) - (((self.df['pY1'] * (1 - self.df['ps']))
                                                                                 / (self.df['ps']))),
                                  self.df['pY1'])

        # Doubly robust estimator under all untreated
        self.df['dr0'] = np.where(self.df[self._exposure] == 0,
                                  (self.df['pY0']),
                                  ((self.df[self._outcome]) / (1 - self.df['ps']) - (
                                              ((self.df['pY0']) * (self.df['ps']))
                                              / (1 - self.df['ps']))))

        # Generating estimates for the risk difference and risk ratio
        self.risk_difference = np.mean(self.df['dr1']) - np.mean(self.df['dr0'])
        self.risk_ratio = np.mean(self.df['dr1']) / np.mean(self.df['dr0'])

    def summary(self, decimal=4):
        """Prints a summary of the results for the doubly robust estimator.

        Parameters
        ----------
        decimal : int, optional
            Number of decimal places to display in the result
        """
        if (self._fit_exposure_model is False) or (self._fit_exposure_model is False):
            raise ValueError('The exposure and outcome models must be specified before the double robust estimate can '
                             'be generated')

        print('----------------------------------------------------------------------')
        print('Risk Difference: ', round(float(self.risk_difference), decimal))
        print('Risk Ratio: ', round(float(self.risk_ratio), decimal))
        print('----------------------------------------------------------------------')


class AIPW:
    def __init__(self, df, exposure, outcome):
        """AIPW is NO longer maintained. Please use AIPTW instead

        Implementation of augmented inverse probablity weight estimator. This is a simple implementation of AIPW
        and does not account for missing data. It only handles time-fixed confounders and treatments at this point.

        The augment-inverse probability weights are calculated from the following formula

        .. math::

            RD = (\frac{Y_{A=1}*A}{\Pr(A=1)} - \frac{\hat{Y}_1*(A-\Pr(A=1)}{\Pr(A=1)}) - (\frac{Y_{A=0}*A}{\Pr(A=0)
            } - \frac{\hat{Y}_0*(A-\Pr(A=1)}{\Pr(A=0)})
            RR = (\frac{Y_{A=1}*A}{\Pr(A=1)} - \frac{\hat{Y}_1*(A-\Pr(A=1)}{\Pr(A=1)}) / (\frac{Y_{A=0}*A}{\Pr(A=0)
            } - \frac{\hat{Y}_0*(A-\Pr(A=1)}{\Pr(A=0)})

        Confidence intervals should be calculated using non-parametric bootstrapping

        Parameters
        ----------
        df : DataFrame
            Pandas DataFrame object containing all variables of interest
        exposure : str
            Column name of the exposure variable. Currently only binary is supported
        outcome : str
            Column name of the outcome variable. Currently only binary is supported
        """
        warnings.warn("AIPW is no longer maintained. AIPW has been replaced by AIPTW. AIPW will no longer be available"
                      " in version 0.5.0", DeprecationWarning)

        self.df = df.copy()
        if df.dropna().shape[0] != df.shape[0]:
            warnings.warn("There is missing data in the dataset. By default, AIPW will drop all missing data. AIPW will"
                          "fit "+str(df.dropna().shape[0])+' of '+str(df.shape[0])+' observations', UserWarning)
        self.df = df.copy().dropna().reset_index()
        self._exposure = exposure
        self._outcome = outcome
        self._fit_exposure_model = False
        self._fit_outcome_model = False
        self._generated_ci = False
        self.risk_difference = None
        self.risk_ratio = None
        self._exp_model = None
        self._out_model = None

    def exposure_model(self, model, print_results=True):
        """Specify the propensity score / inverse probability weight model. Model used to predict the exposure via a
        logistic regression model

        Parameters
        ----------
        model : str
            Independent variables to predict the exposure. For example, 'var1 + var2 + var3'
        print_results : bool, optional
            Whether to print the fitted model results. Default is True (prints results)
        """
        self._exp_model = self._exposure + ' ~ ' + model
        fitmodel = propensity_score(self.df, self._exp_model, print_results=print_results)
        self.df['ps'] = fitmodel.predict(self.df)
        self._fit_exposure_model = True

    def outcome_model(self, model, print_results=True):
        """Specify the outcome model. Model used to predict the outcome via a logistic regression model

        Parameters
        ----------
        model : str
            Independent variables to predict the outcome. For example, 'var1 + var2 + var3 + var4'
        print_results : bool, optional
            Whether to print the fitted model results. Default is True (prints results)
        """
        self._out_model = self._outcome + ' ~ ' + model
        f = sm.families.family.Binomial()
        log = smf.glm(self._out_model, self.df, family=f).fit()
        if print_results:
            print('\n----------------------------------------------------------------')
            print('MODEL: ' + self._out_model)
            print('-----------------------------------------------------------------')
            print(log.summary())

        dfx = self.df.copy()
        dfx[self._exposure] = 1
        self.df['pY1'] = log.predict(dfx)
        dfx = self.df.copy()
        dfx[self._exposure] = 0
        self.df['pY0'] = log.predict(dfx)
        self._fit_outcome_model = True

    def fit(self):
        """Once the exposure and outcome models are specified, we can estimate the Risk Ratio and Risk Difference.
        This function generates the estimated risk difference and risk ratio. For confidence intervals, a non-parametric
        bootstrap procedure should be used
        """
        if (self._fit_exposure_model is False) or (self._fit_outcome_model is False):
            raise ValueError('The exposure and outcome models must be specified before the doubly robust estimate can '
                             'be generated')

        # Doubly robust estimator under all treated
        self.df['dr1'] = np.where(self.df[self._exposure] == 1,
                                  ((self.df[self._outcome]) / self.df['ps']) - (((self.df['pY1'] * (1 - self.df['ps']))
                                                                                 / (self.df['ps']))),
                                  self.df['pY1'])

        # Doubly robust estimator under all untreated
        self.df['dr0'] = np.where(self.df[self._exposure] == 0,
                                  (self.df['pY0']),
                                  ((self.df[self._outcome]) / (1 - self.df['ps']) - (
                                              ((self.df['pY0']) * (self.df['ps']))
                                              / (1 - self.df['ps']))))

        # Generating estimates for the risk difference and risk ratio
        self.risk_difference = np.mean(self.df['dr1']) - np.mean(self.df['dr0'])
        self.risk_ratio = np.mean(self.df['dr1']) / np.mean(self.df['dr0'])

    def summary(self, decimal=4):
        """Prints a summary of the results for the doubly robust estimator.

        Parameters
        ----------
        decimal : int, optional
            Number of decimal places to display in the result
        """
        if (self._fit_exposure_model is False) or (self._fit_exposure_model is False):
            raise ValueError('The exposure and outcome models must be specified before the double robust estimate can '
                             'be generated')

        print('----------------------------------------------------------------------')
        print('Risk Difference: ', round(float(self.risk_difference), decimal))
        print('Risk Ratio: ', round(float(self.risk_ratio), decimal))
        print('----------------------------------------------------------------------')



import warnings
import numpy as np
import pandas as pd
import statsmodels.api as sm
import statsmodels.formula.api as smf
from statsmodels.genmod.families import links


class TimeFixedGFormula:
    def __init__(self, df, exposure, outcome, exposure_type='binary', outcome_type='binary', weights=None):
        """Time-fixed implementation of the g-formula, also referred to as the g-computation algorithm formula. This
        implementation has three options for the treatment courses:

        Currently, only supports binary or continuous outcomes. For binary outcomes a logistic regression
        model to predict probabilities of outcomes via statsmodels. For continuous outcomes a linear regression
        model is used to predict outcomes.
        Binary and multivariate exposures are supported. For binary exposures, a string object of the column name for
        the exposure of interest should be provided. For multivariate exposures, a list of string objects corresponding
        to disjoint indicator terms for the exposure should be provided. Multivariate exposures require the user to
        custom specify treatments when fitting the g-formula. A list of the custom treatment must be provided and be
        the same length as the number of disjoint indicator columns. See
        http://zepid.readthedocs.io/en/latest/ for examples (highly recommended)

        Key options for treatments
            all     -all individuals are given treatment
            none    -no individuals are given treatment
            custom treatments
                    -create a custom treatment. When specifying this, the dataframe must be referred to as 'g' The
                    following is an example that selects those whose age is 30 or younger and are females;
                    treatment="((g['age0']<=30) & (g['male']==0))

        Parameters
        ----------
        df : DataFrame
            Pandas dataframe containing the variables of interest
        exposure : str, list
            Column name for exposure variable label or a list of disjoint indicator exposures
        outcome : str
            Column name for outcome variable
        outcome_type : str, optional
            Outcome variable type. Currently only 'binary', 'normal', and 'poisson variable types are supported
        weights : str, optional
            Column name for weights. Default is None, which assumes every observations has the same weight (i.e. 1)

        Examples
        --------
        Setting up the environment
        >>>from zepid import load_sample_data, spline
        >>>from zepid.causal.gformula import TimeFixedGFormula
        >>>df = load_sample_data(timevary=False)
        >>>df[['cd4_rs1', 'cd4_rs2']] = spline(df, 'cd40', n_knots=3, term=2, restricted=True)
        >>>df[['age_rs1', 'age_rs2']] = spline(df, 'age0', n_knots=3, term=2, restricted=True)

        G-formula with a binary treatment and outcome
        >>>g = TimeFixedGFormula(df, exposure='art', outcome='dead')
        >>>g.outcome_model(model='art + male + age0 + age_rs1 + age_rs2 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')

        >>># Return the estimated marginal outcome under treat-all
        >>>g.fit(treatment='all')
        >>>g.marginal_outcome

        >>># Return the estimated marginal outcome under treat-none
        >>>g.fit(treatment='all')
        >>>g.marginal_outcome

        >>># Return the estimated marginal outcome under custom treatment (treat all females under 40)
        >>>g.fit(treatment="((g['male']==0) & (g['age0']<=40))")
        >>>g.marginal_outcome

        G-formula with a categorical treatment and binary outcome
        >>># Creating categorical variable for CD4 count
        >>>df['cd4_1'] = np.where(((df['cd40'] >= 200) & (df['cd40'] < 400)), 1, 0)
        >>>df['cd4_2'] = np.where(df['cd40'] >= 400, 1, 0)

        >>>g = TimeFixedGFormula(df,exposure=['art_male', 'art_female'], outcome='dead', exposure_type='categorical')
        >>>g.outcome_model(model='cd4_1 + cd4_2 + art + male + age0 + age_rs1 + age_rs2 + dvl0')

        >>># Return marginal outcome under all in reference category (CD4 < 200)
        >>>g.fit(treatment=["False", "False"])

        >>># Return marginal outcome under all in category 1 (CD4 >= 200 & CD4 < 400)
        >>>g.fit(treatment=["True", "False"])

        >>># Return marginal outcome under all in category 2 (CD4 > 400)
        >>>g.fit(treatment=["False", "True"])

        G-formula with binary exposure and continuous (normal-distributed) outcome
        >>>g = TimeFixedGFormula(df,exposure='art', outcome='cd4', outcome_type='normal')
        >>>g.outcome_model(model='art + male + age0 + age_rs1 + age_rs2 + dvl0  + cd40 + cd4_rs1 + cd4_rs2')

        G-formula with binary exposure and continuous (Poisson-distributed) outcome
        >>>g = TimeFixedGFormula(df,exposure='art', outcome='cd4', outcome_type='poisson')
        >>>g.outcome_model(model='art + male + age0 + age_rs1 + age_rs2 + dvl0  + cd40 + cd4_rs1 + cd4_rs2')

        References
        ----------
        Snowden, Jonathan M., Sherri Rose, and Kathleen M. Mortimer. "Implementation of G-computation on a simulated
        data set: demonstration of a causal inference technique." American Journal of Epidemiology 173.7 (2011):
        731-738.
        """
        self.gf = df.copy()
        self.exposure = exposure
        self.outcome = outcome

        if (outcome_type == 'binary') or (outcome_type == 'normal') or (outcome_type == 'poisson'):
            self.outcome_type = outcome_type
        else:
            raise ValueError('Only binary or continuous outcomes are currently supported. Please specify "binary" '
                             '"normal", or "poisson"')

        if (exposure_type == 'binary') or (exposure_type == 'continuous') or (exposure_type == 'categorical'):
            self.exposure_type = exposure_type
        else:
            raise ValueError('Only binary or continuous exposures are currently supported. Please specify "binary", '
                             '"categorical", or "continuous".')

        self._weights = weights
        self._outcome_model = None
        self.marginal_outcome = np.nan
        self.predicted_df = None

    def outcome_model(self, model, print_results=True):
        """Build the model for the outcome. This is also referred to at the Q-model. This must be specified
        before the fit function. If it is not, an error will be raised.

        Parameters
        ----------
        model : str
            Variables to include in the model for predicting the outcome. Must be contained within the input
            pandas dataframe when initialized. Model form should contain the exposure, i.e. 'art + age + male'
        print_results : bool, optional
            Whether to print the logistic regression results to the terminal. Default is True
        """
        if self.outcome_type == 'binary':
            linkdist = sm.families.family.Binomial()
        elif self.outcome_type == 'normal':
            linkdist = sm.families.family.Gaussian()
        else:
            linkdist = sm.families.family.Poisson()

        # Modeling the outcome
        if self._weights is None:
            m = smf.glm(self.outcome+' ~ '+model, self.gf, family=linkdist)
            self._outcome_model = m.fit()
        else:
            m = smf.gee(self.outcome+' ~ '+model, self.gf.index, self.gf, family=linkdist,
                        weights=self.gf[self._weights])
            self._outcome_model = m.fit()

        # Printing results of the model and if any observations were dropped
        if print_results:
            print(self._outcome_model.summary())

    def fit(self, treatment):
        """Fit the parametric g-formula as specified. Binary and multivariate treatments are available. This
        implementation has three options for the binary treatment courses. For multivariate treatments, the user must
        specify custom treatment plans.

        To obtain the confidence intervals, use a bootstrap. See online documentation for an example:
        http://zepid.readthedocs.io/en/latest/

        Parameters
        ----------
        treatment : str, list
            There are three options available for treatment plans. All, none, or a custom pattern
            * all     -all individuals are given treatment
            * none    -no individuals are given treatment
            * custom  -create a custom treatment. When specifying this, the dataframe must be referred to as 'g' The
              following is an example that selects those whose age is 25 or older and are females;
              treatment="((g['age0']>=25) & (g['male']==0))

        Returns
        -------
        marginal_outcome
            Parameter of marginal outcome is filled, which is the mean predicted Y under the treatment strategy
        """
        if self._outcome_model is None:
            raise ValueError('Before the g-formula can be calculated, the outcome model must be specified')
        if (type(treatment) != str) and (type(treatment) != list):
            raise ValueError('Specified treatment must be a string object or a list of string objects')

        # Setting outcome as blank
        g = self.gf.copy()

        # Setting treatment (either multivariate or binary)
        if self.exposure_type == 'binary':
            if type(treatment) == list:
                raise ValueError('A binary exposure is specified. Treatment plan should be a string object')
            if treatment == 'all':
                g[self.exposure] = 1
            elif treatment == 'none':
                g[self.exposure] = 0
            else:  # custom exposure pattern
                g[self.exposure] = np.where(eval(treatment), 1, 0)

        elif self.exposure_type == 'categorical':
            if (treatment == 'all') or (treatment == 'none'):  # Check to make sure custom treatment
                raise ValueError('A multivariate exposure has been specified. A custom treatment must be '
                                 'specified by the user')
            else:
                if len(self.exposure) != len(treatment):  # Check to make sure same about of treatments specified
                    raise ValueError('The list of custom treatment conditions must be the same size as the number of '
                                     'treatments')
                for i in range(len(self.exposure)):
                    g[self.exposure[i]] = np.where(eval(treatment[i]), 1, 0)

                if np.sum(np.where(g[self.exposure].sum(axis=1) > 1, 1, 0)) > 1:
                    warnings.warn('It looks like your specified treatment strategy results in some individuals '
                                  'receiving at least two exposures. Reconsider how the custom treatments are '
                                  'specified', UserWarning)

        else:
            raise ValueError('Still working on allowing for continuous treatments...')
            # TODO fill in this part of continuous exposures

        # Getting predictions
        g[self.outcome] = np.nan
        g[self.outcome] = self._outcome_model.predict(g)
        if self._weights is None:  # unweighted marginal estimate
            self.marginal_outcome = np.mean(g[self.outcome])
        else:  # weighted marginal estimate
            self.marginal_outcome = np.average(g[self.outcome], weights=self.gf[self._weights])
        self.predicted_df = g



import math
import warnings
import patsy
import numpy as np
import statsmodels.api as sm
import statsmodels.formula.api as smf
from scipy.stats import logistic, norm

from zepid.causal.ipw import propensity_score
from zepid.calc import probability_to_odds


class TMLE:
    def __init__(self, df, exposure, outcome, measure='risk_difference', alpha = 0.05):
        """Implementation of a single time-point target maximum likelihood estimator. It uses standard logistic
        regression models to calculate Psi. The TMLE estimator allows a standard logistic regression to be used.
        Alternatively, users are able to directly input predicted outcomes from other methods (like machine learning
        algorithms).

        TMLE is a double robust substitution estimator. For further details, see the listed references. The following
        is a general description of the estimation process for TMLE

        1. Initial estimates for Y are predicted from a regression model. Expected values for each individual are
        generated under the scenarios of all treated vs all untreated

        .. math::

            E(Y|A, L)

        2. Predicted probabilities are generated from a regression model

        .. math::

            \Pr(A=1|L)

        3. The predicted Y is merged together with the IPW using

        .. math::

            H_a(A=a,L) = \frac{I(A=1)}{\pi_1} - \frac{I(A=0)}{\pi_0}

        for each individual. Afterwards, the predicted Y is set as an offset in the following logit model and used to
        predict values under each treatment strategy after fitted

        .. math::

            logit(E(Y|A,L)) = logit(Y_a) + \sigma * H_a

        4. The targeted Psi is estimated via one of the following formulas

        .. math::

            RD = \frac{1}{n} \sum_{i=1}^{n} (\hat{Y}_1 - \hat{Y}_0)

            RR = \frac{1}{n} \sum_{i=1}^{n} (\hat{Y}_1) / \frac{1}{n} \sum_{i=1}^{n} (\hat{Y}_0)

        Confidence intervals are constructed using influence curve theory

        Parameters
        ----------
        df : DataFrame
            Pandas dataframe containing the variables of interest
        exposure : str
            Column label for the exposure of interest
        outcome : str
            Column label for the outcome of interest
        measure : str, optional
            What the TMLE psi estimates. Current options include; risk difference comparing treated to untreated
            (F(A=1) - F(A=0)), risk ratio (F(A=1) / F(A=0)), and odds ratio. The following keywords are available
            * 'risk_difference'  :   F(A=1) - F(A=0)
            * 'risk_ratio'       :   F(A=1) / F(A=0)
            * 'odds_ratio'       :   (F(A=1) / (1 - F(A=1)) / (F(A=0) / (1 - F(A=0))
        alpha : int, optional
            Alpha for confidence interval level. Default is 0.05

        Examples
        --------
        Setting up environment
        >>>from zepid import load_sample_data, spline
        >>>from zepid.causal.doublyrobust import TMLE
        >>>df = load_sample_data(False).dropna()
        >>>df[['cd4_rs1', 'cd4_rs2']] = spline(df, 'cd40', n_knots=3, term=2, restricted=True)

        Estimating TMLE using logistic regression
        >>>tmle = TMLE(df, exposure='art', outcome='dead')
        >>># Specifying exposure/treatment model
        >>>tmle.exposure_model('male + age0 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')
        >>># Specifying outcome model
        >>>tmle.exposure_model('art + male + age0 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')
        >>># TMLE estimation procedure
        >>>tmle.fit()
        >>># Printing main results
        >>>tmle.summary()
        >>># Extracting point estimate and confidence intervals, respectively
        >>>tmle.psi
        >>>tmle.confint

        Estimating TMLE with machine learning algorithm from sklearn
        >>>from sklearn.linear_model import LogisticRegression
        >>>log1 = LogisticRegression(penalty='l1', random_state=201)
        >>>tmle = TMLE(df, 'art', 'dead')
        >>># custom_model allows specification of machine learning algorithms
        >>>tmle.exposure_model('male + age0 + cd40 + cd4_rs1 + cd4_rs2 + dvl0', custom_model=log1)
        >>>tmle.outcome_model('male + age0 + cd40 + cd4_rs1 + cd4_rs2 + dvl0', custom_model=log1)
        >>>tmle.fit()

        Demonstration of estimating g-model with symmetric bounds
        >>>tmle.exposure_model('male + age0 + cd40 + cd4_rs1 + cd4_rs2 + dvl0', bound=0.05)

        Demonstration of estimating g-model with asymmetric bounds
        >>>tmle.exposure_model('male + age0 + cd40 + cd4_rs1 + cd4_rs2 + dvl0', bound=[0.05, 0.9])

        References
        ----------
        Schuler, Megan S., and Sherri Rose. "Targeted maximum likelihood estimation for causal inference in
        observational studies." American journal of epidemiology 185.1 (2017): 65-73.

        Van der Laan, Mark J., and Sherri Rose. Targeted learning: causal inference for observational and experimental
        data. Springer Science & Business Media, 2011.

        Van Der Laan, Mark J., and Daniel Rubin. "Targeted maximum likelihood learning." The International Journal of
        Biostatistics 2.1 (2006).
        """
        if df.dropna().shape[0] != df.shape[0]:
            warnings.warn("There is missing data in the dataset. By default, TMLE will drop all missing data. TMLE will"
                          "fit "+str(df.dropna().shape[0])+' of '+str(df.shape[0])+' observations', UserWarning)

        # Detailed steps follow "Targeted Learning" chapter 4, figure 4.2 by van der Laan, Rose
        self._psi_correspond = measure
        self.df = df.copy().dropna().reset_index()
        self.alpha = alpha
        self._exposure = exposure
        self._outcome = outcome
        self._out_model = None
        self._exp_model = None
        self._fit_exposure_model = False
        self._fit_outcome_model = False
        self.QA0W = None
        self.QA1W = None
        self.QAW = None
        self.g1W = None
        self.g0W = None
        self._epsilon = None
        self.psi = None
        self.confint = None

    def exposure_model(self, model, custom_model=None, bound=False, print_results=True):
        """Estimation of Pr(A=1|L), which is termed as g(A=1|L) in the literature

        Parameters
        ----------
        model : str
            Independent variables to predict the exposure. Example) 'var1 + var2 + var3'
        custom_model : optional
            Input for a custom model that is used in place of the logit model (default). The model must have the
            "fit()" and  "predict()" attributes. Both sklearn and supylearner are supported as custom models. In the
            background, TMLE will fit the custom model and generate the predicted probablities
        bound : float, list, optional
            Value between 0,1 to truncate predicted probabilities. Helps to avoid near positivity violations.
            Specifying this argument can improve finite sample performance for random positivity violations. However,
            inference becomes limited to the restricted population. Default is False, meaning no truncation of
            predicted probabilities occurs. Providing a single float assumes symmetric trunctation. A collection of
            floats can be provided for asymmetric trunctation
        print_results : bool, optional
            Whether to print the fitted model results. Default is True (prints results)
        """
        self._exp_model = self._exposure + ' ~ ' + model

        # Step 3) Estimation of g-model (exposure model)
        if custom_model is None:
            fitmodel = propensity_score(self.df, self._exp_model, print_results=print_results)
            self.g1W = fitmodel.predict(self.df)

        # User-specified prediction model
        else:
            data = patsy.dmatrix(model + ' - 1', self.df)
            try:
                fm = custom_model.fit(X=data, y=self.df[self._outcome])
            except TypeError:
                raise TypeError("Currently custom_model must have the 'fit' function with arguments 'X', 'y'. This "
                                "covers both sklearn and supylearner. If there is a predictive model you would "
                                "like to use, please open an issue at https://github.com/pzivich/zepid and I "
                                "can work on adding support")
            if print_results and hasattr(fm, 'summarize'):
                fm.summarize()
            if hasattr(fm, 'predict_proba'):
                self.g1W = fm.predict_proba(data)[:, 1]
            elif hasattr(fm, 'predict'):
                self.g1W = fm.predict(data)
            else:
                raise ValueError("Currently custom_model must have 'predict' or 'predict_proba' attribute")

        self.g0W = 1 - self.g1W
        if bound:  # Bounding predicted probabilities if requested
            self.g1W = self._bounding(self.g1W, bounds=bound)
        if bound:  # Bounding predicted probabilities if requested
            self.g0W = self._bounding(self.g0W, bounds=bound)

        self._fit_exposure_model = True

    def outcome_model(self, model, custom_model=None, print_results=True):
        """Estimation of E(Y|A,L), which is also written sometimes as Q(A,W) or Pr(Y=1|A,W)

        Parameters
        ----------
        model : str
            Independent variables to predict the exposure. Example) 'var1 + var2 + var3'
        custom_model : optional
            Input for a custom model that is used in place of the logit model (default). The model must have the
            "fit()" and  "predict()" attributes. Both sklearn and supylearner are supported as custom models. In the
            background, TMLE will fit the custom model and generate the predicted probablities
        print_results : bool, optional
            Whether to print the fitted model results. Default is True (prints results)
        """
        self._out_model = self._outcome + ' ~ ' + model

        # Step 1) Prediction for Q (estimation of Q-model)
        if custom_model is None:  # Logistic Regression model for predictions
            f = sm.families.family.Binomial()
            log = smf.glm(self._out_model, self.df, family=f).fit()

            if print_results:
                print('\n----------------------------------------------------------------')
                print('MODEL: ' + self._out_model)
                print('-----------------------------------------------------------------')
                print(log.summary())

            # Step 2) Estimation under the scenarios
            self.QAW = log.predict(self.df)
            dfx = self.df.copy()
            dfx[self._exposure] = 1
            self.QA1W = log.predict(dfx)
            dfx = self.df.copy()
            dfx[self._exposure] = 0
            self.QA0W = log.predict(dfx)

        # User-specified model
        else:
            data = patsy.dmatrix(model + ' - 1', self.df)
            try:
                fm = custom_model.fit(X=data, y=self.df[self._outcome])
            except TypeError:
                raise TypeError("Currently custom_model must have the 'fit' function with arguments 'X', 'y'. This "
                                "covers both sklearn and supylearner. If there is a predictive model you would "
                                "like to use, please open an issue at https://github.com/pzivich/zepid and I "
                                "can work on adding support")
            if print_results and hasattr(fm, 'summarize'):
                fm.summarize()
            if hasattr(fm, 'predict_proba'):
                self.QAW = fm.predict_proba(data)[:, 1]

                dfx = self.df.copy()
                dfx[self._exposure] = 1
                data = patsy.dmatrix(model + ' - 1', dfx)
                self.QA1W = fm.predict_proba(data)[:, 1]

                dfx = self.df.copy()
                dfx[self._exposure] = 0
                data = patsy.dmatrix(model + ' - 1', dfx)
                self.QA0W = fm.predict_proba(data)[:, 1]

            elif hasattr(fm, 'predict'):
                self.QAW = fm.predict(data)

                dfx = self.df.copy()
                dfx[self._exposure] = 1
                data = patsy.dmatrix(model + ' - 1', dfx)
                self.QA1W = fm.predict(data)

                dfx = self.df.copy()
                dfx[self._exposure] = 0
                data = patsy.dmatrix(model + ' - 1', dfx)
                self.QA0W = fm.predict(data)

            else:
                raise ValueError("Currently custom_model must have 'predict' or 'predict_proba' attribute")

        self._fit_outcome_model = True

    def fit(self):
        """Estimates psi based on the gAW and QAW. Confidence intervals come from the influence curve

        Returns
        -------
        TMLE gains Psi and confint attributes
        """
        if (self._fit_exposure_model is False) or (self._fit_outcome_model is False):
            raise ValueError('The exposure and outcome models must be specified before the psi estimate can '
                             'be generated')

        # Step 4) Calculating clever covariate (HAW)
        H1W = self.df[self._exposure] / self.g1W
        H0W = -(1 - self.df[self._exposure]) / (self.g0W)
        HAW = H1W + H0W

        # Step 5) Estimating TMLE
        f = sm.families.family.Binomial()
        log = sm.GLM(self.df[self._outcome], np.column_stack((H1W, H0W)), offset=np.log(probability_to_odds(self.QAW)),
                     family=f).fit()
        self._epsilon = log.params
        Qstar1 = logistic.cdf(np.log(probability_to_odds(self.QA1W)) + self._epsilon[0] * 1/self.g1W)
        Qstar0 = logistic.cdf(np.log(probability_to_odds(self.QA0W)) + self._epsilon[1] * -1/self.g0W)
        Qstar = log.predict(np.column_stack((H1W, H0W)), offset=np.log(probability_to_odds(self.QAW)))

        # Step 6) Calculating Psi
        if self.alpha == 0.05:  # Without this, won't match R exactly. R relies on 1.96, while I use SciPy
            zalpha = 1.96
        else:
            zalpha = norm.ppf(1 - self.alpha / 2, loc=0, scale=1)

        # p-values are not implemented (doing my part to enforce CL over p-values)
        if self._psi_correspond == 'risk_difference':
            self.psi = np.mean(Qstar1 - Qstar0)
            # Influence Curve for CL
            ic = HAW * (self.df[self._outcome] - Qstar) + (Qstar1 - Qstar0) - self.psi
            varIC = np.var(ic, ddof=1) / self.df.shape[0]
            self.confint = [self.psi - zalpha * math.sqrt(varIC),
                            self.psi + zalpha * math.sqrt(varIC)]

        elif self._psi_correspond == 'risk_ratio':
            self.psi = np.mean(Qstar1) / np.mean(Qstar0)
            # Influence Curve for CL
            ic = (1/np.mean(Qstar1)*(H1W * (self.df[self._outcome] - Qstar) + Qstar1 - np.mean(Qstar1)) -
                  (1/np.mean(Qstar0))*(-1*H0W * (self.df[self._outcome] - Qstar) + Qstar0 - np.mean(Qstar0)))
            varIC = np.var(ic, ddof=1) / self.df.shape[0]
            self.confint = [np.exp(np.log(self.psi) - zalpha * math.sqrt(varIC)),
                            np.exp(np.log(self.psi) + zalpha * math.sqrt(varIC))]

        elif self._psi_correspond == 'odds_ratio':
            self.psi = (np.mean(Qstar1) / (1 - np.mean(Qstar1))) / (np.mean(Qstar0) / (1 - np.mean(Qstar0)))
            # Influence Curve for CL
            ic = ((1/(np.mean(Qstar1)*(1-np.mean(Qstar1))) * (H1W*(self.df[self._outcome] - Qstar) + Qstar1)) -
                  (1/(np.mean(Qstar0)*(1 - np.mean(Qstar0))) * (-1*H0W*(self.df[self._outcome] - Qstar) + Qstar0)))
            seIC = math.sqrt(np.var(ic, ddof=1) / self.df.shape[0])
            self.confint = [np.exp(np.log(self.psi) - zalpha * seIC),
                            np.exp(np.log(self.psi) + zalpha * seIC)]

        else:
            raise ValueError('Specified parameter is not implemented. Please use one of the available options')

    def summary(self, decimal=3):
        """Prints summary of model results

        Parameters
        ----------
        decimal : int, optional
            Number of decimal places to display. Default is 3
        """
        if (self._fit_exposure_model is False) or (self._fit_exposure_model is False):
            raise ValueError('The exposure and outcome models must be specified before the psi estimate can '
                             'be generated')

        print('----------------------------------------------------------------------')
        print('Psi: ', round(float(self.psi), decimal))
        print(str(round(100 * (1 - self.alpha), 1)) + '% two-sided CI: (' + str(round(self.confint[0], decimal)), ',',
              str(round(self.confint[1], decimal)) + ')')
        print('----------------------------------------------------------------------')
        print('Psi corresponds to '+self._psi_correspond)
        print('----------------------------------------------------------------------')

    @staticmethod
    def _bounding(v, bounds):
        """Background function to perform bounding feature. Not intended for users to access

        v:
            -Values to be bounded
        bounds:
            -Percentile thresholds for bounds
        """
        if type(bounds) is float:  # Symmetric bounding
            if bounds < 0 or bounds > 1:
                raise ValueError('Bound value must be between (0, 1)')
            v = np.where(v < bounds, bounds, v)
            v = np.where(v > 1-bounds, 1-bounds, v)
        elif type(bounds) is str:  # Catching string inputs
            raise ValueError('Bounds must either be a float between (0, 1), or a collection of floats between (0, 1)')
        else:  # Asymmetric bounds
            if bounds[0] > bounds[1]:
                raise ValueError('Bound thresholds must be listed in ascending order')
            if len(bounds) > 2:
                warnings.warn('It looks like your specified bounds is more than two floats. Only the first two '
                              'specified bounds are used by the bound statement. So only ' +
                              str(bounds[0:2]) + ' will be used', UserWarning)
            if type(bounds[0]) is str or type(bounds[1]) is str:
                raise ValueError('Bounds must be floats between (0, 1)')
            if (bounds[0] < 0 or bounds[1] > 1) or (bounds[0] < 0 or bounds[1] > 1):
                raise ValueError('Both bound values must be between (0, 1)')
            v = np.where(v < bounds[0], bounds[0], v)
            v = np.where(v > bounds[1], bounds[1], v)
        print(np.max(v))
        return v




import warnings
import numpy as np
import pandas as pd
import statsmodels.api as sm
import statsmodels.formula.api as smf
from statsmodels.genmod.families import links


class TimeVaryGFormula:
    def __init__(self, df, idvar, exposure, outcome, time_out, time_in=None, method='MonteCarlo', weights=None):
        """Time-varying implementation of the g-formula, also referred to as the g-computation algorithm formula. The
        time-varying parametric g-formula uses either the Monte Carlo or the sequential regression (iterative
        expectations) estimators. The Monte Carlo estimator is useful for survival data and the sequential regression
        estimator is useful for longitudinal data. This implementation has four options for the treatment courses:

        Options for treatments
        * all : all individuals are given treatment
        * none : no individuals are given treatment
        * natural : individuals retain their observed treatment
        * custom : create a custom treatment. When specifying this, the dataframe must be  referred to as 'g' The following
            is an example that selects those whose age is 25 or older and are females
            Ex) treatment="((g['age0']>=25) & (g['male']==0))

        Currently, only binary exposures and a binary outcomes are supported. Logistic regression models are used to
        predict exposures and outcomes via statsmodels. See Keil et al. (2014) for a good description of the
        time-varying g-formula. See http://zepid.readthedocs.io/en/latest/ for an example (highly recommended)

        Parameters
        ----------
        df : DataFrame
            Pandas dataframe containing the variables of interest
        exposure : str
            Treatment column label
        outcome : str
            Outcome column label
        time_out : str
            End of follow-up period time column label
        time_in : str, optional
            Start of follow-up period time label. This column is only required for the Monte Carlo estimator
        method : str, optional
            Estimator to use estimate the g-formula. Default is the Monte Carlo estimator. The Monte Carlo estimator is
            requested via 'MonteCarlo' and the sequential regression (iterative conditional) is requested with
            'SequentialRegression'
        weights : str, optional
            Column label for weights. Default is None, which assumes every observations has the same weight (i.e. 1)

        Notes
        -----
        1) Monte Carlo increases by time units of one. Input dataset should reflect this
        2) Only binary exposures and binary outcomes are supported
        3) Binary and continuous covariates are supported
        4) The labeling of the covariate models is important. They are fit in the order that they are labeled!
        5) For the Monte Carlo estimator, fit the natural course model first and compare to the observed data. They should
           be similar

        Process for the Monte Carlo g-formula
            1) run lines in "in_recode"
            2) time-varying covariates, order ascending in from "labels"
                a) predict time-varying covariate
                b) run lines in "recode" from "add_covariate_model()"
            3) predict exposure / apply exposure pattern
            4) predict outcome
            5) run lines in "out_recode"
            6) lag variables in "lags"
            7) append current-time rows to full dataframe
            8) Repeat till t_max is met

        Process for the sequential regression g-formula
            1) Convert long data set to wide data set
            2) Identify individuals who followed the counterfactual treatment plan and had the outcome
            3) Fit a regression model for the outcome at time t for Y
            4) Predict outcomes under the observed treatment and the counterfactual treatment
            5) Repeat regression model fitting for t-1 to min(t)
            6) Take the mean predicted Y at the end to obtain the cumulative probability

        Examples
        --------
        Setting up the environment
        >>>import numpy as np
        >>>from zepid import load_sample_data, spline
        >>>from zepid.causal.gformula import TimeVaryGFormula
        >>>df = load_sample_data(timevary=True)
        >>>df['lag_art'] = df['art'].shift(1)
        >>>df['lag_art'] = np.where(df.groupby('id').cumcount() == 0, 0, df['lag_art'])
        >>>df['lag_cd4'] = df['cd4'].shift(1)
        >>>df['lag_cd4'] = np.where(df.groupby('id').cumcount() == 0, df['cd40'], df['lag_cd4'])
        >>>df['lag_dvl'] = df['dvl'].shift(1)
        >>>df['lag_dvl'] = np.where(df.groupby('id').cumcount() == 0, df['dvl0'], df['lag_dvl'])
        >>>df[['age_rs0', 'age_rs1', 'age_rs2']] = spline(df, 'age0', n_knots=4, term=2, restricted=True)
        >>>df['cd40_sq'] = df['cd40'] ** 2
        >>>df['cd40_cu'] = df['cd40'] ** 3
        >>>df['cd4_sq'] = df['cd4'] ** 2
        >>>df['cd4_cu'] = df['cd4'] ** 3
        >>>df['enter_sq'] = df['enter'] ** 2
        >>>df['enter_cu'] = df['enter'] ** 3

        Estimating the g-formula with the Monte Carlo estimator
        >>>g = TimeVaryGFormula(df, idvar='id', exposure='art', outcome='dead', time_in='enter', time_out='out')

        >>># Specifying the exposure/treatment model
        >>>exp_m = 'male + age0 + age_rs0 + age_rs1 + age_rs2 + cd40 + cd40_sq + cd40_cu + dvl0 + cd4 + cd4_sq +' +
        >>>        'cd4_cu + dvl + enter + enter_sq + enter_cu'
        >>>g.exposure_model(exp_m, restriction="g['lag_art']==0")  # restriction enforces intent-to-treat

        >>># Specifying the outcome model
        >>>out_m = 'art + male + age0 + age_rs0 + age_rs1 + age_rs2 + cd40 + cd40_sq + cd40_cu + dvl0 + cd4 +' +
        >>>        'cd4_sq + cd4_cu + dvl + enter + enter_sq + enter_cu'
        >>>g.outcome_model(out_m, restriction="g['drop']==0")  # restriction enforces loss-to-follow-up

        >>># Specifying the time-varying confounder models
        >>>dvl_m = 'male + age0 + age_rs0 + age_rs1 + age_rs2 + cd40 + cd40_sq + cd40_cu + dvl0 + lag_cd4 + ' +
        >>>        'lag_dvl + lag_art + enter + enter_sq + enter_cu'
        >>>g.add_covariate_model(label=1, covariate='dvl', model=dvl_m, var_type='binary')
        >>>cd4_m = 'male + age0 + age_rs0 + age_rs1 + age_rs2 +  cd40 + cd40_sq + cd40_cu + dvl0 + lag_cd4 + ' +
        >>>        'lag_dvl + lag_art + enter + enter_sq + enter_cu'
        >>>cd4_recode_scheme = ("g['cd4'] = np.maximum(g['cd4'],1);"  # Recode scheme makes sure variables are recoded
        >>>                     "g['cd4_sq'] = g['cd4']**2;"
        >>>                     "g['cd4_cu'] = g['cd4']**3")
        >>>g.add_covariate_model(label=2, covariate='cd4', model=cd4_m, recode=cd4_recode_scheme, var_type='continuous')

        >>># Estimating outcomes under a simulated Markov Chain Monte Carlo for natural course
        >>>g.fit(treatment="((g['art']==1) | (g['lag_art']==1))",  # Treatment plan (natural course in this case)
        >>>      lags={'art': 'lag_art',  # Creating variables to lag in the process
        >>>            'cd4': 'lag_cd4',
        >>>            'dvl': 'lag_dvl'},
        >>>      sample=50000,  # Number of resamples to use (should be large number to reduce simulation error)
        >>>      t_max=None,  # Maximum time to simulate to (None uses data set maximum time)
        >>>      in_recode=("g['enter_sq'] = g['enter']**2;"
        >>>                 "g['enter_cu'] = g['enter']**3"))  # How to recode time in each time-step
        >>># See website documentation for further instructions
        >>># (https://zepid.readthedocs.io/en/latest/Causal.html#g-computation-algorithm-monte-carlo)

        Estimating the g-formula with the Sequential Regression (Interative Conditional) estimator
        >>># Setting up data
        >>>from zepid import load_longitudinal_data
        >>>df = load_longitudinal_data()
        >>>g = TimeVaryGFormula(data, idvar='id', exposure='A', outcome='Y', time_out='t',
        >>>                     method='SequentialRegression')  # Specify for sequential regression estimator

        >>># Specify the outcome model
        >>>g.outcome_model('A + L', print_results=False)

        >>># Estimate the marginal risk under treat-all for max(t)
        >>>g.fit(treatment="all")
        >>>print(g.predicted_outcomes)

        >>># Estimate the marginal risk under treat-all for t=2
        >>>g.fit(treatment="all", t_max=2)
        >>>print(g.predicted_outcomes)

        >>># Estimate the marginal risk under a custom treatment (treat all except at time 2)
        >>>g.fit(treatment="g['t'] != 2")
        >>>print(g.predicted_outcomes)

        References
        ----------
        Keil, A.P., Edwards, J.K., Richardson, D.B., Naimi, A.I., Cole, S.R. (2014). The Parametric g-Formula for Time-
        to-Event Data: Intuition and a Worked Example. Epidemiology 25(6), 889-897

        Kreif, N., Tran, L., Grieve, R., De Stavola, B., Tasker, R. C., & Petersen, M. (2017). Estimating the
        comparative effectiveness of feeding interventions in the pediatric intensive care unit: a demonstration of
        longitudinal targeted maximum likelihood estimation. American Journal of Epidemiology, 186(12), 1370-1379.
        """
        self.gf = df.copy()
        self.idvar = idvar

        # Checking that treatment is binary
        if df[exposure].dropna().value_counts().index.isin([0, 1]).all():
            self.exposure = exposure
        else:
            raise ValueError('Only binary treatments/exposures are currently implemented')

        # Checking that outcome is binary
        if df[outcome].dropna().value_counts().index.isin([0, 1]).all():
            self.outcome = outcome
        else:
            raise ValueError('Only binary outcomes are currently implemented')

        self.exp_model = None
        self.out_model = None
        self.time_in = time_in
        self.time_out = time_out
        self._exposure_model_fit = False
        self._outcome_model_fit = False
        self._covariate_models = []
        self._covariate_model_index = []
        self._covariate = []
        self._covariate_type = []
        self._covariate_recode = []
        self._weights = weights
        self.predicted_outcomes = None

        # Different Estimator Approaches
        if method == 'MonteCarlo':
            self._mc = True
            if time_in is None:
                raise ValueError('"time_in" must be specified for MonteCarlo estimation')
        elif method == 'SequentialRegression':
            self._mc = False
            self._modelform = None
            self._printseqregresults = None
        else:
            raise ValueError('Either "MonteCarlo" or "SequentialRegression" must be specified as the estimation '
                             'procedure for TimeVaryGFormula')

    def exposure_model(self, model, restriction=None, print_results=True):
        """Add a specified regression model for the exposure. This is used for natural course estimation of the Monte
        Carlo g-formula. If the Monte Carlo estimation procedure is used, this must be specified before calling the
        fit function.

        Parameters
        ----------
        model : str
            Variables to include in the model for predicting the exposure. Must be contained within the input
            pandas dataframe when initialized. Format follows patsy standards
            For example) 'var1 + var2 + var3 + var4'
        restriction : str, optional
            Used to restrict the population that the regression model is fit to. Useful for Intent-to-Treat model
            fitting. The pandas dataframe must be referred to as 'g'. For example) "g['art']==1"
        print_results : bool, optional
            Whether to print the logistic regression model results to the terminal. Default is True
        """
        g = self.gf.copy()
        if restriction is not None:
            g = g.loc[eval(restriction)].copy()
        linkdist = sm.families.family.Binomial()

        if self._weights is None:  # Unweighted g-formula
            self.exp_model = smf.glm(self.exposure + ' ~ ' + model, g, family=linkdist).fit()
        else:  # Weighted g-formula
            self.exp_model = smf.gee(self.exposure + ' ~ ' + model, self.idvar, g, weights=g[self._weights],
                                     family=linkdist).fit()

        if print_results:
            print(self.exp_model.summary())
        self._exposure_model_fit = True

    def outcome_model(self, model, restriction=None, print_results=True):
        """Add a specified regression model for the outcome. This is used for both Monte Carlo and sequential regression
        estimators and must be specified before the fit function.

        Parameters
        ----------
        model:
            Variables to include in the model for predicting the outcome. Must be contained within the input
            pandas dataframe when initialized. Format follows patsy standards
            For example) 'var1 + var2 + var3 + var4'
        restriction : str, optional
            Used to restrict the population that the regression model is fit to. Useful for Intent-to-Treat model
            fitting. The pandas dataframe must be referred to as 'g'. For example) "g['art']==1"
        print_results : bool, optional
            Whether to print the logistic regression model results to the terminal. Default is True
        """
        if self._mc:
            g = self.gf.copy()
            if restriction is not None:
                g = g.loc[eval(restriction)].copy()
            linkdist = sm.families.family.Binomial()

            if self._weights is None:  # Unweighted g-formula
                self.out_model = smf.glm(self.outcome + ' ~ ' + model, g, family=linkdist).fit()
            else:  # Weighted g-formula
                self.out_model = smf.gee(self.outcome + ' ~ ' + model, self.idvar, g, weights=g[self._weights],
                                         family=linkdist).fit()

            if print_results:
                print(self.out_model.summary())
        else:
            self._modelform = model
            self._printseqregresults = print_results

        self._outcome_model_fit = True

    def add_covariate_model(self, label, covariate, model, restriction=None, recode=None, var_type='binary',
                            print_results=True):
        """Add a specified regression model for time-varying confounders. Unlike the exposure and outcome models, a
        covariate model does NOT have to be specified. Additionally, *n* covariate models can be specified for *n*
        time-varying covariates. Additional models are added by repeated calls for this function with the corresponding
        covariates and predictive regression equations

        This argument is only used for the Monte Carlo g-formula. The sequential regression only requires specification
        of the outcome model.

        Parameters
        ----------
        label : int
            Integer label for the covariate model. Covariate models are fit in ascending order within
             TimeVaryGFormula
        covariate : str
            Column label for time-varying confounder to be predicted
        model : str
            Variables to include in the model for predicting the outcome. Must be contained within the input
            pandas dataframe when initialized. Format follows patsy
            For example) 'var1 + var2 + var3 + var4'
        restriction : str, optional
            Used to restrict the population to fit the logistic regression model to. Useful for Intent-to-Treat
            model fitting. The pandas dataframe must be referred to as 'g'. For example) "g['art']==1"
        recode : str, optional
            This variable is vitally important for various functional forms implemented later in models. This
            is used to run some background code to recreate functional forms as the g-formula is estimated via fit()
            For an example, let's say we have age but we want the functional form to be quadratic. For this, we
            would set the recode="g['age_sq'] = g['age']**2;" Similar to TimeFixedGFormula, 'g' must be specified as the
            DataFrame object with the corresponding indexes. Also lines of executable code should end with ';', so
            Python knows that the line ends there. My apologies for this poor solution... I am working on a better way.
            In the background, Python executes the code input into recode
        var_type : str, optional
            Type of variable that the covariate is. Current options include 'binary' or 'continuous'
        print_results : bool, optional
            Whether to print the logistic regression model results to the terminal. Default is True
        """
        if type(label) is not int:
            raise ValueError('Label must be an integer')

        # Building predictive model
        g = self.gf.copy()
        if restriction is not None:
            g = g.loc[eval(restriction)].copy()

        if self._weights is None:  # Unweighted g-formula
            if var_type == 'binary':
                linkdist = sm.families.family.Binomial()
                m = smf.glm(covariate + ' ~ ' + model, g, family=linkdist)
            elif var_type == 'continuous':
                linkdist = sm.families.family.Gaussian(sm.families.links.identity)
                m = smf.gls(covariate + ' ~ ' + model, g)
            else:
                raise ValueError('Only binary or continuous covariates are currently supported')
        else:  # Weighted g-formula
            if var_type == 'binary':
                linkdist = sm.families.family.Binomial()
                m = smf.gee(covariate + ' ~ ' + model, self.idvar, g, weights=g[self._weights], family=linkdist)
            elif var_type == 'continuous':
                linkdist = sm.families.family.Gaussian(sm.families.links.identity)
                m = smf.gee(covariate + ' ~ ' + model, self.idvar, g, weights=g[self._weights], family=linkdist)
            else:
                raise ValueError('Only binary or continuous covariates are currently supported')

        f = m.fit()
        if print_results:
            print(f.summary())

        # Adding to lists, it is used to predict variables later on for the time-varying...
        self._covariate_models.append(f)
        self._covariate_model_index.append(label)
        self._covariate.append(covariate)
        self._covariate_type.append(var_type)
        if recode is None:
            self._covariate_recode.append('None')  # Must be string for exec() to use later
        else:
            self._covariate_recode.append(recode)

    def fit(self, treatment, lags=None, sample=10000, t_max=None, in_recode=None, out_recode=None):
        """Estimate the counterfactual outcomes under the specified treatment plan using the previously specified
        regression models. For the Monte Carlo g-formula, both the exposure and outcome models need to be specified
        before fit can be called. For sequential regression, only the outcome model needs to be specified.

        Parameters
        ----------
        treatment : str
            Treatment strategy. Options include
            * all : all individuals are given treatment
            * none : no individuals are given treatment
            * natural : individuals retain their observed treatment. Only available for MonteCarlo
            * custom : create a custom treatment. When specifying this, the dataframe must be  referred to as 'g'
                The following is an example that selects those whose age is 25 or older and are females
                Ex) treatment="((g['age0']>=25) & (g['male']==0))
        lags : dict, optional
            Dictionary of variable names and the corresponding lagged variable name. This should be specified for all
            variables that are lagged. This parameter is only used for the Monte Carlo g-formula.
             As an example, {'art':'lagged_art'} would correctly lag the variable 'art' to be 'lagged_art' for each time
             step of the Monte Carlo procedure
        sample : int, optional
            Number of individuals to sample from the original data with replacement. This argument is only used by the
            Monte Carlo g-formula. The number of samples to use should be a large number to reduce simulation error.
            The default is 10000
        t_max : int, optional
            Maximum time to run Monte Carlo g-formula until. Default is None, which uses the maximum time of the input
            dataframe.
        in_recode : str, optional
            On the fly recoding of variables done before the Monte Carlo loop starts. Needed to do any kind of
            functional forms for entry times. This is executed at each start of the Monte Carlo g-formula time steps
        out_recode : str, optional
            On the fly recoding of variables done at the end of the Monte Carlo loop. Needed for operations like
            counting the number of days with a treatment. This is executed at each end of the Monte Carlo g-formula
            time steps
        """
        if self._outcome_model_fit is False:
            raise ValueError('Before the g-formula can be calculated, the outcome model must be specified')
        if (type(treatment) != str) and (type(treatment) != list):
            raise ValueError('Specified treatment must be a string object')

        # Monte Carlo Estimation
        if self._mc:
            if self._exposure_model_fit is False:
                raise ValueError('Before the g-formula can be calculated, the exposure model must be specified for '
                                 'Monte Carlo estimation')
            # Getting data all set
            if self._weights is None:
                gs = self.gf.loc[(self.gf.groupby(self.idvar).cumcount() == 0) == True].sample(n=sample, replace=True)
            else:
                gs = self.gf.loc[(self.gf.groupby(self.idvar).cumcount() == 0) == True].sample(n=sample,
                                                                                               weights=self._weights,
                                                                                               replace=True)
            gs['uid_g_zepid'] = [v for v in range(sample)]

            # Background preparations
            gs[self.outcome] = 0

            # getting maximum time steps to run g-formula
            if t_max is None:
                t_max = np.max(self.gf[self.time_out])

            # Estimating via MC process
            self.predicted_outcomes = self._monte_carlo(gs, treatment, t_max, in_recode, out_recode, lags)

        # Sequential Regression Estimation
        else:
            if self._exposure_model_fit or len(self._covariate_models) > 0:
                raise ValueError('Only the outcome model needs be specified for the sequential regression estimator')
            self.predicted_outcomes = self._sequential_regression(treatment=treatment, tmax=t_max)

    def _monte_carlo(self, gs, treatment, t_max, in_recode, out_recode, lags):
        """Hidden function that executes the Monte Carlo estimation process for the g-formula
        """
        # setting up some parts outside of Monte Carlo loop to speed things up
        mc_simulated_data = []
        g = gs.copy()
        if len(self._covariate_models) != 0:
            cov_model_order = (sorted(range(len(self._covariate_model_index)),
                                      key=self._covariate_model_index.__getitem__))
            run_cov = True
        else:
            run_cov = False

        # Monte Carlo for loop
        for i in range(int(t_max)):
            g = g.loc[g[self.outcome] == 0].reset_index(drop=True).copy()
            g[self.time_in] = i
            if in_recode is not None:
                exec(in_recode)

            # predict time-varying covariates
            if run_cov:
                for j in cov_model_order:
                    g[self._covariate[j]] = self._predict(df=g, model=self._covariate_models[j],
                                                          variable=self._covariate_type[j])
                    exec(self._covariate_recode[j])

            # predict exposure when customized treatments
            if treatment == 'all':
                g[self.exposure] = 1
            elif treatment == 'none':
                g[self.exposure] = 0
            elif treatment == 'natural':
                g[self.exposure] = self._predict(df=g, model=self.exp_model, variable='binary')
            else:  # custom exposure pattern
                g[self.exposure] = self._predict(df=g, model=self.exp_model, variable='binary')
                g[self.exposure] = np.where(eval(treatment), 1, 0)

            # predict outcome
            g[self.outcome] = self._predict(df=g, model=self.out_model, variable='binary')
            g[self.time_out] = i + 1

            # executing any code before appending
            if out_recode is not None:
                exec(out_recode)

            # updating lagged variables
            if lags is not None:
                for k, v in lags.items():
                    g[v] = g[k]

            # stacking simulated data in a list
            mc_simulated_data.append(g)

        try:
            gs = pd.concat(mc_simulated_data, ignore_index=True, sort=False)
        except TypeError:  # gets around pandas <0.22 error
            gs = pd.concat(mc_simulated_data, ignore_index=True)
        if self._weights is None:
            return gs[['uid_g_zepid', self.exposure, self.outcome, self.time_in, self.time_out] +
                      self._covariate].sort_values(by=['uid_g_zepid',
                                                       self.time_in]).reset_index(drop=True)
        else:
            return gs[['uid_g_zepid', self.exposure, self.outcome, self._weights, self.time_in,
                       self.time_out] + self._covariate].sort_values(by=['uid_g_zepid',
                                                                         self.time_in]).reset_index(drop=True)

    def _sequential_regression(self, treatment, tmax):
        """Hidden function that executes the sequential regression estimation for g-formula
        """
        # TODO allow option to include different estimation models for each time point or the same model
        if treatment == 'natural':
            # Thoughts: MC estimator needs natural course as a check. This should not apply to SR estimator
            raise ValueError('Natural course estimation is not clear to me with Sequential Regression Estimator. '
                             'Therefore, "natural" is not implemented')

        # If custom treatment, it gets evaluated here
        g = self.gf
        if treatment not in ['all', 'none']:
            g['__indicator'] = np.where(eval(treatment), 1, 0)

        # Restricting based on tmax argument
        if tmax is None:
            pass
        elif tmax in list(self.gf[self.time_out].unique()):
            g = g.loc[g[self.time_out] <= tmax].copy()
        else:
            warnings.warn("The t_max argument specifies a time that is not observed in the data. All times less than"
                          "the specified t_max argument included in the estimation procedure", UserWarning)
            g = g.loc[g[self.time_out] <= tmax].copy()

        # Converting dataframe from long-to-wide for easier estimation
        column_labels = list(g.columns)  # Getting all column labels (important to match with formula)
        df = self._long_to_wide(df=g, id=self.idvar, t=self.time_out)
        linkdist = sm.families.family.Binomial()
        rt_points = sorted(list(g[self.time_out].unique()), reverse=True)  # Getting all t's to backward loop
        t_points = sorted(list(g[self.time_out].unique()), reverse=False)  # Getting all t's to forward loop

        # Checking for recurrent outcomes. Recurrent are not currently supported
        if pd.Series(df[[self.outcome + '_' + str(t) for t in sorted(t_points, reverse=False)
                         ]].sum(axis=1, skipna=True) > 1).any():
            raise ValueError('Looks like your data has multiple outcomes. Recurrent outcomes are not currently '
                             'supported')

        # Step 1: Creating indicator for individuals who followed counterfactual outcome
        treat_t_points = []
        for t in t_points:
            # Following treatment strategy
            # alternative: if treat all, can do simple multiplication. if treat none, can do (1-A) simple multiplication
            if treatment == 'all':
                df['__indicator_' + str(t)] = np.where(df[self.exposure + '_' + str(t)] == 0, 0, np.nan)
                df['__indicator_' + str(t)] = np.where(df[self.exposure + '_' + str(t)] == 1, 1,
                                                       df['__indicator_' + str(t)])
            elif treatment == 'none':
                df['__indicator_' + str(t)] = np.where(df[self.exposure + '_' + str(t)] == 0, 1, np.nan)
                df['__indicator_' + str(t)] = np.where(df[self.exposure + '_' + str(t)] == 1, 0,
                                                       df['__indicator_' + str(t)])
            else:  # custom exposure pattern
                pass

            treat_t_points.append('__indicator_' + str(t))
            df['__check_' + str(t)] = df[treat_t_points + [self.outcome + '_' + str(t)]].prod(axis=1, skipna=True)

            # This following check carries forward the outcome under the counterfactual treatment
            if t_points.index(t) == 0:
                pass
            else:
                df['__check_' + str(t)] = np.where(df['__check_' + str(t_points[t_points.index(t) - 1])] == 1,
                                                   1, df['__check_' + str(t)])

        # Step 2: Sequential Regression Estimation
        for t in rt_points:
            # 2.1) Relabel everything to match with the specified model (selecting out that timepoint is within)
            d_labels = {}
            for c in column_labels:
                d_labels[c + '_' + str(t)] = c
            g = df.filter(regex='_' + str(t)).rename(mapper=d_labels, axis=1).reset_index().copy()
            g[self.time_out] = t

            # 2.2) Fit the model to the observed data
            if rt_points.index(t) == 0:
                if self._weights is None:
                    m = smf.glm(self.outcome + ' ~ ' + self._modelform, g, family=linkdist).fit()  # GLM
                else:
                    m = smf.gee(self.outcome + ' ~ ' + self._modelform, self.idvar, g,
                                weights=df[self._weights + '_' + str(t)], family=linkdist).fit()  # Weighted, so GEE
                if self._printseqregresults:
                    print(m.summary())
            else:
                # Uses previous predicted values to estimate
                g[self.outcome] = np.where(df['__pred_'+self.outcome+'_'+str(t_points[t_points.index(t)+1])].isna(),
                                           g[self.outcome],
                                           df['__pred_' + self.outcome + '_' + str(t_points[t_points.index(t)+1])])

                if self._weights is None:
                    m = smf.glm(self.outcome + ' ~ ' + self._modelform, g, family=linkdist).fit()  # GLM
                else:
                    m = smf.gee(self.outcome + ' ~ ' + self._modelform, self.idvar, g,
                                weights=df[self._weights + '_' + str(t)], family=linkdist).fit()  # Weighted, so GEE
                if self._printseqregresults:
                    print(m.summary())

            # 2.3) Getting Counterfactual Treatment Values
            if treatment == 'all':
                g[self.exposure] = 1
            elif treatment == 'none':
                g[self.exposure] = 0
            else:
                g[self.exposure] = np.where(eval(treatment), 1, 0)

            # Predicted values based on counterfactual treatment strategy from predicted model
            df['__pred_' + self.outcome + '_' + str(t)] = np.where(df[self.outcome + '_' + str(t)].isna(),
                                                                   np.nan,
                                                                   m.predict(g))
            # If followed counterfactual treatment & had outcome, then always considered to have outcome past that t
            df['__cf_' + self.outcome + '_' + str(t)] = np.where((df['__check_' + str(t)] == 1),
                                                                 1,
                                                                 df['__pred_' + self.outcome + '_' + str(t)])

        # Step 3) Returning estimated results
        if self._weights is None:
            return np.mean(df['__pred_' + self.outcome + '_' + str(t_points[0])])
        else:
            return np.average(df['__pred_' + self.outcome + '_' + str(t_points[0])],
                              weights=df[self._weights + '_' + str(t_points[0])])

    @staticmethod
    def _predict(df, model, variable):
        """Hidden predict method to shorten the Monte Carlo estimation code
        """
        # pp = data.mul(model.params).sum(axis=1) # Alternative to statsmodels.predict(), but too much too implement
        pp = model.predict(df)
        if variable == 'binary':
            # pp = odds_to_probability(np.exp(pp))  # assumes a logit model. For non-statsmodel.predict() option
            pred = np.random.binomial(1, pp, size=len(pp))
        elif variable == 'continuous':
            pred = np.random.normal(loc=pp, scale=np.std(model.resid), size=len(pp))
        else:
            raise ValueError('That option is not supported')
        return pred

    @staticmethod
    def _long_to_wide(df, id, t):
        """Hidden function for sequential regression that converts from long to wide data set
        """
        reshaped = []
        for c in df.columns:
            if c == id or c == t:
                pass
            else:
                df['v'] = c + '_' + df[t].astype(str)
                reshaped.append(df.pivot(index='id', columns='v', values=c))

        return pd.concat(reshaped, axis=1)




import numpy as np


def trapezoidal(mini, mode1, mode2, maxi, size=None):
    """Creates trapezoidal distribution based on Fox & Lash 2005. This function can be used to generate distributions
    of probabilities and effect measures for sensitivity analyses. It is particularly useful when used in conjunction
    with rr_corr to determine a distribution of potential results due to a single unadjusted
    confounder

    Parameters
    --------------
    mini : float
        Minimum value of trapezoidal distribution
    mode1 : float
        Start of uniform distribution
    mode2 : float
        End of uniform distribution
    maxi : float
        Maximum value of trapezoidal distribution
    size : int, optional
        Number of observations to generate. Default is None, which returns a single draw

    Returns
    --------------
    float or array
        Returns either a single float from the distribution or an array of floats

    Examples
    --------------
    Single draw from a trapezoidal distribution
    >>>from zepid.sensitivity_analysis import trapezoidal
    >>>trapezoidal(mini=0.2, mode1=0.3, mode2=0.5, maxi=0.6)

    100 draws from a trapezoidal distribution
    >>>trapezoidal(mini=0.2, mode1=0.3, mode2=0.5, maxi=0.6, size=100)
    """
    if size is None:
        p = np.random.uniform()
        v = (p*(maxi+mode2-mini-mode1)+(mini+mode1)) / 2
        if v < mode1:
            v = mini + np.sqrt((mode1-mini)*(2*v-mini-mode1))
        elif v > mode2:
            v = maxi - np.sqrt(2*(maxi-mode2)*(v-mode2))
        else:
            pass   
        return v
    
    # Draws the number specified by size
    elif type(size) is int:
        va = []
        for i in range(size):
            va.append(trapezoidal(mini=mini, mode1=mode1, mode2=mode2, maxi=maxi, size=None))
        return np.array(va)
    
    # ValueError is size is not an integer
    else:
        raise ValueError('"size" must be an integer')




from setuptools import setup

exec(compile(open('zepid/version.py').read(),
             'zepid/version.py', 'exec'))

setup(name='zepid',
      version=__version__,
      description='Tool package for epidemiologic analyses',
      keywords='epidemiology inverse-probability-weights risk-ratio g-computation g-formula IPW AIPW TMLE',
      packages=['zepid',
                'zepid.calc',
                'zepid.graphics',
                'zepid.sensitivity_analysis',
                'zepid.causal.ipw',
                'zepid.causal.gformula',
                'zepid.causal.doublyrobust',
                'zepid.datasets'],
      include_package_data=True,
      license='MIT',
      author='Paul Zivich',
      author_email='zepidpy@gmail.com',
      url='https://github.com/pzivich/zepid',
      classifiers=['Programming Language :: Python :: 3.5',
                   'Programming Language :: Python :: 3.6'],
      install_requires=['pandas>=0.18',
                        'numpy',
                        'statsmodels>=0.7.0',
                        'matplotlib>=2.0',
                        'scipy',
                        'tabulate',
                        'patsy'],
      )



"""
Contains useful graphic generators. Currently, effect measure plots and functional form assessment plots
are implemented. Uses matplotlib to generate graphics. Future inclusions include forest plots

Contents
----------------
Functional form assessment- func_form_plot()
Forest plot/ effect measure plot- EffectMeasurePlot()
P-value distribution plot- pvalue_plot()
Spaghetti plot- spaghetti_plot()
Receiver-Operator Curve- roc()
Dynamic risk plot- dynamic_risk_plot()
"""

import warnings
import numpy as np
import pandas as pd
from scipy.stats import norm
import statsmodels.api as sm
import statsmodels.formula.api as smf
from statsmodels.genmod.families import links
from statsmodels.nonparametric.smoothers_lowess import lowess
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.ticker as mticker


class EffectMeasurePlot:
    """Used to generate effect measure plots. effectmeasure plot accepts four list type objects.
    effectmeasure_plot is initialized with the associated names for each line, the point estimate,
    the lower confidence limit, and the upper confidence limit.

    Notes
    ----------
    .. code::

            _____________________________________________      Measure     % CI
            |                                           |
        1   |        --------o-------                   |       x        n, 2n
            |                                           |
        2   |                   ----o----               |       w        m, 2m
            |                                           |
            |___________________________________________|
                #           #           #           #

    Examples
    -------------
    Setting up the data to plot
    >>> from zepid.graphics import EffectMeasurePlot
    >>>lab = ['One','Two']
    >>>emm = [1.01,1.31]
    >>>lcl = ['0.90',1.01]
    >>>ucl = [1.11,1.53]

    Setting up the plot, measure labels, and point colors
    >>>x = EffectMeasurePlot(lab, emm, lcl, ucl)
    >>>x.labels(effectmeasure='RR')  # Changing label of measure
    >>>x.colors(pointcolor='r')  # Changing color of the points

    Generating matplotlib axes object of forest plot
    >>>x.plot(t_adjuster=0.13)
    """
    def __init__(self, label, effect_measure, lcl, ucl):
        """Initializes effectmeasure_plot with desired data to plot. All lists should be the same
        length. If a blank space is desired in the plot, add an empty character object (' ') to
        each list at the desired point.

        Parameters
        --------------
        label : list
            List of labels to use for y-axis
        effect_measure : list
            List of numbers for point estimates to plot. If point estimate has trailing zeroes,
            input as a character object rather than a float
        lcl : list
            List of numbers for upper confidence limits to plot. If point estimate has trailing
            zeroes, input as a character object rather than a float
        ucl : list
            List of numbers for upper confidence limits to plot. If point estimate has
            trailing zeroes, input as a character object rather than a float
        """
        self.df = pd.DataFrame()
        self.df['study'] = label
        self.df['OR'] = effect_measure
        self.df['LCL'] = lcl
        self.df['UCL'] = ucl
        self.df['OR2'] = self.df['OR'].astype(str).astype(float)
        if (all(isinstance(item, float) for item in lcl)) & (all(isinstance(item, float) for item in effect_measure)):
            self.df['LCL_dif'] = self.df['OR'] - self.df['LCL']
        else:
            self.df['LCL_dif'] = (pd.to_numeric(self.df['OR'])) - (pd.to_numeric(self.df['LCL']))
        if (all(isinstance(item, float) for item in ucl)) & (all(isinstance(item, float) for item in effect_measure)):
            self.df['UCL_dif'] = self.df['UCL'] - self.df['OR']
        else:
            self.df['UCL_dif'] = (pd.to_numeric(self.df['UCL'])) - (pd.to_numeric(self.df['OR']))
        self.em = 'OR'
        self.ci = '95% CI'
        self.scale = 'linear'
        self.center = 1
        self.errc = 'dimgrey'
        self.shape = 'd'
        self.pc = 'k'
        self.linec = 'gray'

    def labels(self, **kwargs):
        """Function to change the labels of the outputted table. Additionally, the scale and reference
        value can be changed.

        Parameters
        -------------
        effectmeasure : string, optional
            Changes the effect measure label
        conf_int : string, optional
            Changes the confidence interval label
        scale : string, optional
            Changes the scale to either log or linear
        center : float, integer, optional
            Changes the reference line for the center
        """
        if 'effectmeasure' in kwargs:
            self.em = kwargs['effectmeasure']
        if 'conf_int' in kwargs:
            self.ci = kwargs['conf_int']
        if 'scale' in kwargs:
            self.scale = kwargs['scale']
        if 'center' in kwargs:
            self.center = kwargs['center']

    def colors(self, **kwargs):
        """Function to change colors and shapes.

        Parameters
        ---------------
        errorbarcolor : string, optional
            Changes the error bar colors
        linecolor : string, optional
            Changes the color of the reference line
        pointcolor : string, optional
            Changes the color of the points
        pointshape : string, optional
            Changes the shape of points
        """
        if 'errorbarcolor' in kwargs:
            self.errc = kwargs['errorbarcolor']
        if 'pointshape' in kwargs:
            self.shape = kwargs['pointshape']
        if 'linecolor' in kwargs:
            self.linec = kwargs['linecolor']
        if 'pointcolor' in kwargs:
            self.pc = kwargs['pointcolor']

    def plot(self, figsize=(3, 3), t_adjuster=0.01, decimal=3, size=3, max_value=None, min_value=None):
        """Generates the matplotlib effect measure plot with the default or specified attributes.
        The following variables can be used to further fine-tune the effect measure plot

        Parameters
        -----------------
        figsize : tuple, optional
            Adjust the size of the figure. Syntax is same as matplotlib `figsize`
        t_adjuster : float, optional
            Used to refine alignment of the table with the line graphs. When generate plots, trial and error for this
            value are usually necessary. I haven't come up with an algorithm to determine this yet...
        decimal : integer, optional
            Number of decimal places to display in the table
        size : integer,
            Option to adjust the size of the lines and points in the plot
        max_value : float, optional
            Maximum value of x-axis scale. Default is None, which automatically determines max value
        min_value : float, optional
            Minimum value of x-axis scale. Default is None, which automatically determines min value

        Returns
        ---------
        matplotlib axes
        """
        tval = []
        ytick = []
        for i in range(len(self.df)):
            if not np.isnan(self.df['OR2'][i]):
                if ((isinstance(self.df['OR'][i], float)) & (isinstance(self.df['LCL'][i], float)) &
                        (isinstance(self.df['UCL'][i], float))):
                    tval.append([round(self.df['OR2'][i], decimal), (
                                '(' + str(round(self.df['LCL'][i], decimal)) + ', ' +
                                str(round(self.df['UCL'][i], decimal)) + ')')])
                else:
                    tval.append(
                        [self.df['OR'][i], ('(' + str(self.df['LCL'][i]) + ', ' + str(self.df['UCL'][i]) + ')')])
                ytick.append(i)
            else:
                tval.append([' ', ' '])
                ytick.append(i)
        if max_value is None:
            if pd.to_numeric(self.df['UCL']).max() < 1:
                maxi = round(((pd.to_numeric(self.df['UCL'])).max() + 0.05),
                             2)  # setting x-axis maximum for UCL less than 1
            if (pd.to_numeric(self.df['UCL']).max() < 9) and (pd.to_numeric(self.df['UCL']).max() >= 1):
                maxi = round(((pd.to_numeric(self.df['UCL'])).max() + 1),
                             0)  # setting x-axis maximum for UCL less than 10
            if pd.to_numeric(self.df['UCL']).max() > 9:
                maxi = round(((pd.to_numeric(self.df['UCL'])).max() + 10),
                             0)  # setting x-axis maximum for UCL less than 100
        else:
            maxi = max_value
        if min_value is None:
            if pd.to_numeric(self.df['LCL']).min() > 0:
                mini = round(((pd.to_numeric(self.df['LCL'])).min() - 0.1), 1)  # setting x-axis minimum
            if pd.to_numeric(self.df['LCL']).min() < 0:
                mini = round(((pd.to_numeric(self.df['LCL'])).min() - 0.05), 2)  # setting x-axis minimum
        else:
            mini = min_value
        plt.figure(figsize=figsize)  # blank figure
        gspec = gridspec.GridSpec(1, 6)  # sets up grid
        plot = plt.subplot(gspec[0, 0:4])  # plot of data
        tabl = plt.subplot(gspec[0, 4:])  # table of OR & CI
        plot.set_ylim(-1, (len(self.df)))  # spacing out y-axis properly
        if self.scale == 'log':
            try:
                plot.set_xscale('log')
            except:
                raise ValueError('For the log scale, all values must be positive')
        plot.axvline(self.center, color=self.linec, zorder=1)
        plot.errorbar(self.df.OR2, self.df.index, xerr=[self.df.LCL_dif, self.df.UCL_dif], marker='None', zorder=2,
                      ecolor=self.errc, elinewidth=(size / size), linewidth=0)
        plot.scatter(self.df.OR2, self.df.index, c=self.pc, s=(size * 25), marker=self.shape, zorder=3,
                     edgecolors='None')
        plot.xaxis.set_ticks_position('bottom')
        plot.yaxis.set_ticks_position('left')
        plot.get_xaxis().set_major_formatter(matplotlib.ticker.ScalarFormatter())
        plot.get_xaxis().set_minor_formatter(matplotlib.ticker.NullFormatter())
        plot.set_yticks(ytick)
        plot.set_xlim([mini, maxi])
        plot.set_xticks([mini, self.center, maxi])
        plot.set_xticklabels([mini, self.center, maxi])
        plot.set_yticklabels(self.df.study)
        plot.yaxis.set_ticks_position('none')
        plot.invert_yaxis()  # invert y-axis to align values properly with table
        tb = tabl.table(cellText=tval, cellLoc='center', loc='right', colLabels=[self.em, self.ci],
                        bbox=[0, t_adjuster, 1, 1])
        tabl.axis('off')
        tb.auto_set_font_size(False)
        tb.set_fontsize(12)
        for key, cell in tb.get_celld().items():
            cell.set_linewidth(0)
        return plot


def functional_form_plot(df, outcome, var, f_form=None, outcome_type='binary', discrete=False, link_dist=None,
                         loess=True, loess_value=0.4, legend=True, model_results=True, points=False):
    """Creates a functional form plot to aid in functional form assessment for continuous/discrete variables. Plots can
    be created for binary and continuous outcomes. Default options are set to create a functional form plot for a
    binary outcome. To convert to a continuous outcome, outcome_type needs to be changed, in addition to the link_dist

    Parameters
    ------------
    df : DataFrame
        Pandas dataframe that contains the variables of interest
    outcome : string
        Column name of the outcome variable of interest
    var : string
        Column name of the variable of interest for the functional form assessment
    f_form : string, optional
        Regression equation of the functional form to assess. Default is None, which will produce a linear functional
        form. Input the regression equation as the variables of interest, separated by +. For example, 'var + var_sq'
    outcome_type : string, optional
        Variable type of the outcome variable. Currently, only binary and continuous variables are
        supported. Default is 'binary' but 'continuous' is also supported
    link_dist : optional
        Link and distribution for the GLM regression equation. Change this to any valid link and distributions
        supported by statsmodels. Default is None, which conducts logistic regression
    loess_value : float, optional
        Fraction of observations to use to fit the LOESS curve. This will need to be changed iteratively to determine
        which percent works best for the data. Default is 0.4
    legend : bool, optional
        Turn the legend on or off. Default is True, displaying the legend in the graph
    model_results : bool, optional
        Whether to produce the model results. Default is True, which provides model results
    loess : bool, optional
        Whether to plot the LOESS curve along with the functional form. Default is True
    points : bool, optional
        Whether to plot the data points, where size is relative to the number of observations. Default is False
    discrete : bool, optional
        If your data is truly continuous, leave setting to bin the dat. Will automatically bin observations into
        categories for generation of the LOESS curve. If you data is discrete, you can set this to True to use your
        actual values. If you get a perfect SeparationError from statsmodels, it means you might have to reshift your
        categories.

    Returns
    -----------
    matplotlib axes
        Returns a matplotlib graph with a LOESS line (dashed red-line), regression line (sold blue-line), and
        confidence interval (shaded blue)

    Examples
    ------------
    Setting up the environment
    >>>from zepid import load_sample_data
    >>>from zepid.graphics import functional_form_plot
    >>>import matplotlib.pyplot as plt
    >>>df = load_sample_data(timevary=False)
    >>>df['cd4_sq'] = df['cd4']**2

    Creating a functional form plot for a linear functional form
    >>>functional_form_plot(df, outcome='dead', var='cd4')
    >>>plt.show()

    Functional form assessment for a quadractic functional form
    >>>functional_form_plot(df, outcome='dead', var='cd4', f_form='cd4 + cd4_sq')
    >>>plt.show()

    Varying the LOESS value (increased LOESS value to smooth LOESS curve further)
    >>>functional_form_plot(df, outcome='dead', var='cd4', loess_value=0.5)
    >>>plt.show()

    Removing the LOESS curve and the legend from the plot
    >>>functional_form_plot(df, outcome='dead', var='cd4', loess=False, legend=False)
    >>>plt.show()

    Adding summary points to the plot. Points are grouped together and their size reflects their relative n
    >>>functional_form_plot(df, outcome='dead', var='cd4', loess=False, legend=False, points=True)
    >>>plt.show()

    Functional form assessment for a discrete variable (age)
    >>>functional_form_plot(df, outcome='dead', var='age0', discrete=True)
    >>>plt.show()
    """
    # Copying out the dataframe to a new object we will manipulate a bit
    rf = df.copy()
    rf = rf.dropna(subset=[var, outcome]).sort_values(by=[var, outcome]).reset_index()
    warnings.warn('Warning: missing observations of model variables are dropped. ' +
                  str(int(df.shape[0] - rf.shape[0])) +
                  ' observations were dropped from the functional form assessment')

    # Functional form for the model
    if f_form is None:
        f_form = var
    else:
        pass

    # Generating Models
    if outcome_type == 'binary':
        if link_dist is None:
            link_dist = sm.families.family.Binomial()
        else:
            pass
    elif outcome_type == 'continuous':
        if link_dist is None:
            link_dist = sm.families.family.Gaussian(sm.families.links.identity)
        else:
            pass
    else:
        raise ValueError('Only binary or continuous outcomes are currently supported')

    # Generating LOESS or points if requested
    ax = plt.gca()
    if loess or points:
        if outcome_type == 'binary':
            if discrete is False:
                # Binning continuous variable into categories to get "General" functional form
                categories = int((np.max(rf[var]) - np.min(rf[var])) / 5)
                if model_results:
                    print('''A total of ''' + str(categories) + ''' categories were created. If you would like to 
                            influence  the number of categories the spline is fit to, do the following\n\tIncrease: 
                            multiply by constant >1\n\tDecrease: multiply by contast <1 and >0''')
                rf['vbin'] = pd.qcut(rf[var], q=categories, duplicates='drop').cat.codes
                djm = smf.glm(outcome + '~ C(vbin)', rf, family=link_dist).fit()
            else:
                djm = smf.glm(outcome + '~ C(' + var + ')', rf, family=link_dist).fit()
            djf = djm.get_prediction(rf).summary_frame()
            dj = pd.concat([rf, djf], axis=1)
            dj.sort_values(var, inplace=True)
            if points:
                pf = dj.groupby(by=[var, 'mean']).count().reset_index()
                ax.scatter(pf[var], pf['mean'], s=[100 * (n / np.max(pf[var])) for n in pf[var]],
                           color='gray', label='Data point')
            if loess:
                yl = lowess(list(dj['mean']), list(dj[var]), frac=loess_value)
                lowess_x = list(zip(*yl))[0]
                lowess_y = list(zip(*yl))[1]
                ax.plot(lowess_x, lowess_y, '--', color='red', linewidth=1, label='LOESS')
        if outcome_type == 'continuous':
            if points:
                pf = rf.groupby(by=[var, outcome]).count().reset_index()
                ax.scatter(pf[var], pf[outcome], color='gray', label='Data point')
            if loess:
                yl = lowess(list(rf[outcome]), list(rf[var]), frac=loess_value)
                lowess_x = list(zip(*yl))[0]
                lowess_y = list(zip(*yl))[1]
                ax.plot(lowess_x, lowess_y, '--', color='red', linewidth=1, label='LOESS')

    # Functional form model fitting
    ffm = smf.glm(outcome + ' ~ ' + f_form, rf, family=link_dist).fit()
    if model_results is True:
        print(ffm.summary())
        print('AIC: ', ffm.aic)
        print('BIC: ', ffm.bic)
    fff = ffm.get_prediction(rf).summary_frame()
    ff = pd.concat([rf, fff], axis=1)
    ff.sort_values(var, inplace=True)

    # Generating plot for functional form
    ax.fill_between(ff[var], ff['mean_ci_upper'], ff['mean_ci_lower'], alpha=0.1, color='blue', label='95% CI')
    ax.plot(ff[var], ff['mean'], '-', color='blue', label='Regression')
    ax.set_xlabel(var)
    ax.set_ylabel('Outcome')
    if legend is True:
        ax.legend()
    return ax


def pvalue_plot(point, sd, color='b', fill=True, null=0, alpha=None):
    """Creates a plot of the p-value distribution based on a point estimate and standard deviation.
    I find this plot to be useful to explain p-values and how much evidence weight you have in a
    specific value. I think it is useful to explain what exactly a p-value tells you. Note that this
    plot only works for measures on a linear scale (i.e. it will plot exp(log(RR)) incorrectly). It also
    helps to understand what exactly confidence intervals are telling you. These  plots are based on
    Rothman Epidemiology 2nd Edition pg 152-153 and explained more fully within.

    Parameters
    -------------
    point : float
        Point estimate. Must be on a linear scale (RD / log(RR))
    sd : float
        Standard error of the estimate. Must for linear scale (SE(RD) / SE(log(RR)))
    color : str, optional
        Change color of p-value plot
    fill : bool, optional
        Hhether to fill the curve under the p-value distribution. Setting to False prevents fill
    null : float, integer, optional
        The main value to compare to. The default is zero
    alpha : float, optional
        Whether to draw a line designating significance level area. Default is None, which does not draw this line.
        Generally, would be set to 0.05 to correspond to the widely used alpha of 0.05

    Returns
    -----------
    matplotlib axes

    Examples
    -----------
    Setting up the environment
    >>>from zepid.graphics import pvalue_plot
    >>>import matplotlib.pyplot as plt

    Basic P-value plot
    >>>pvalue_plot(point=-0.1, sd=0.061, color='r')
    >>>plt.show()

    P-value plot with significance line drawn at 'alpha'
    >>>pvalue_plot(point=-0.1, sd=0.061, color='r', alpha=0.025)
    >>>plt.show()

    P-value plot with different comparison value
    >>>pvalue_plot(point=-0.1, sd=0.061, color='r', null=0.1)
    >>>plt.show()
    """
    if point <= null:
        lower = (point - 3 * sd)
        if (point + 3 * sd) < 0:
            upper = point + 3 * sd
        else:
            upper = null + 3 * sd
    if point > null:
        upper = (point + 3 * sd)
        if (point - 3 * sd) > 0:
            lower = null - 3 * sd
        else:
            lower = point - 3 * sd
    ax = plt.gca()
    x1 = np.linspace(lower, point, 100)
    x2 = np.linspace(point, upper, 100)
    ax.plot(x2, 2 * (1 - norm.cdf(x2, loc=point, scale=sd)), c=color)
    ax.plot(x1, 2 * norm.cdf(x1, loc=point, scale=sd), c=color)
    if fill == True:
        ax.fill_between(x2, 2 * (1 - norm.cdf(x2, loc=point, scale=sd)), color=color, alpha=0.2)
        ax.fill_between(x1, 2 * norm.cdf(x1, loc=point, scale=sd), color=color, alpha=0.2)
    ax.vlines(null, 0, 1, colors='k')
    ax.set_xlim([lower, upper])
    ax.set_ylim([0, 1])
    ax.set_ylabel('P-value')
    if alpha is not None:
        ax.hlines(alpha, lower, upper)
    return ax


def spaghetti_plot(df, idvar, variable, time):
    """Create a spaghetti plot by an ID variable. A spaghetti plot can be useful for visualizing
    trends or looking at longitudinal data patterns for individuals all at once.

    Parameters
    ------------
    df : DataFrame
        Pandas dataframe containing variables of interest
    idvar : str
        ID variable for observations. This should indicate the group or individual followed over the time variable
    variable : str
        Variable of interest to see how it varies over time
    time : str
        Time or other variable in which the variable variation occurs

    Returns
    -----------
    matplotlib axes

    Examples
    -----------
    Setting up the environment
    >>>from zepid import load_sample_data
    >>>from zepid.graphics import spaghetti_plot
    >>>df = load_sample_data(timevary=True)

    Generating spaghetti plot for changing CD4 count
    >>>spaghetti_plot(df, idvar='id', variable='cd4', time='enter')
    >>>plt.show()
    """
    ax = plt.gca()
    for i in df[idvar].unique():
        s = df.loc[df[idvar] == i].copy()
        s.sort_values(time, ascending=False)
        ax.plot(s[time], s[variable])
    ax.set_xlabel(time)
    ax.set_ylabel(variable)
    return ax


def roc(df, true, threshold, youden_index=True):
    """Generate a Receiver Operator Curve from true values and predicted probabilities. Youden's Index can also be
    calculated. Youden's index is calculated as

    .. math::

        P_{Yi} = max(Se_i + Sp_i - 1)

    Parameters
    ------------
    df : DataFrame
        Pandas dataframe containing variables of interest
    true : str
        True designation of the outcome (1, 0)
    threshold : str
        Predicted probabilities for the outcome
    youden_index : bool, optional
        Whether to calculate Youden's index. Youden's index maximizes both sensitivity and specificity. The formula
        finds the maximum of (sensitivity + specificity - 1)

    Returns
    -----------
    matplotlib axes
    """
    sens = []
    fpr = []
    thresh = []
    tf = df[[threshold, true]].copy()
    if tf.isnull().values.sum() != 0:
        raise ValueError('ROC curve cannot handle missing data for probability or true values')

    # Getting all possible cutpoints
    values = (list(np.unique(tf[threshold])))
    values = [float(np.min(tf[threshold]) - 0.001)] + values + [float(np.max(tf[threshold]) + 0.001)]
    # Going through all the cutpoints and calculating Sensitivity and 1-Specificity
    for v in list(reversed(values)):
        thresh.append(v)
        prediction = np.where(tf[threshold] >= v, 1, 0)
        se = prediction[tf[true] == 1].mean()
        sens.append(se)
        sp = prediction[tf[true] == 0].mean()
        fpr.append(sp)

    # If requested, calculate Youden's Index
    if youden_index is True:
        spec = [1 - i for i in fpr]
        youdens = []
        for i, j in zip(sens, spec):
            youdens.append(i + j - 1)
        ind = np.argmax(youdens)
        print('----------------------------------------------------------------------')
        print("Youden's Index: ", thresh[ind])
        print("Predictive values at Youden's Index")
        print("\tSensitivity: ", sens[ind])
        print("\tSpecificity: ", spec[ind])
        print('----------------------------------------------------------------------')

    # Creating ROC plot
    ax = plt.gca()
    ax.plot(fpr, sens, color='blue')
    ax.plot([0, 1], [0, 1], color='gray', linestyle='--')
    if youden_index is True:
        ax.text(0.65, 0.35, "Youden's Index:\n      " + str(round(thresh[ind], 5)))
    ax.set_xlim([-0.01, 1.01])
    ax.set_ylim([-0.01, 1.01])
    ax.set_ylabel('Sensitivity')
    ax.set_xlabel('1 -Specificity')
    return ax


def dynamic_risk_plot(risk_exposed, risk_unexposed, measure='RD', loess=True, loess_value=0.25, point_color='darkblue',
                      line_color='b', scale='linear'):
    """Creates a plot of risk measures over time. See Cole et al. "Estimation of standardized risk difference and ratio
    in a competing risks framework: application to injection drug use and progression to AIDS after initiation of
    antiretroviral therapy." Am J Epidemiol. 2015 for an example of this plot

    Parameters
    --------------
    risk_exposed : Series
        Pandas Series with the probability of the outcome among the exposed group. Index by 'timeline' where 'timeline'
        is the time. If you directly output the 1 - survival_function_ from lifelines.KaplanMeierFitter(), this should
        create a valid input
    risk_unexposed : Series
        Pandas Series with the probability of the outcome among the exposed group. Index by 'timeline' where 'timeline'
        is the time
    measure : str, optional
        Whether to generate the risk difference (RD) or risk ratio (RR). Default is 'RD'
    loess : bool, optional
        Whether to generate LOESS curve fit to the calculated points. Default is True
    loess_value : float, optional
        Fraction of values to fit LOESS curve to. Default is 0.25
    point_color : str, optional
        Color of the points
    line_color : str, optional
        Color of the LOESS line generated and plotted
    scale : str, optional
        Change the y-axis scale. Options are 'linear' (default), 'log', 'log-transform'. 'log' and 'log-transform' is
        only a valid option for Risk Ratio plots

    Returns
    -----------
    matplotlib axes

    Examples
    --------
    See graphics documentation or causal documentation
    """
    re = risk_exposed.drop_duplicates(keep='first').iloc[:, 0].rename('exposed').reset_index()
    ru = risk_unexposed.drop_duplicates(keep='first').iloc[:, 0].rename('unexposed').reset_index()
    re.timeline = np.round(re.timeline * 100000).astype(int) # This avoids a merge issue on floats
    ru.timeline = np.round(ru.timeline * 100000).astype(int)
    r = pd.merge(re, ru, how='outer', left_on='timeline', right_on='timeline').sort_values(by='timeline')
    r.timeline /= 100000
    r.ffill(inplace=True)
    if measure == 'RD':
        r['m'] = r['exposed'] - r['unexposed']
    elif measure == 'RR':
        r['m'] = r['exposed'] / r['unexposed']
        if scale == 'log-transform':
            r['m'] = np.log(r['m'])
    else:
        raise ValueError('Only "RD" and "RR" are currently supported')

    # Generating the plot
    ax = plt.gca()
    ax.plot(r['timeline'], r['m'], 'o', c=point_color)
    if loess is True:
        l = lowess(list(r['m']), list(r['timeline']), frac=loess_value)
        lowess_x = list(zip(*l))[0]
        lowess_y = list(zip(*l))[1]
        ax.plot(lowess_x, lowess_y, '-', c=line_color, linewidth=4)
    if measure == 'RD':
        ax.hlines(0, 0, np.max(r['timeline'] + 0.5), linewidth=1.5)
        ax.set_ylabel('Risk Difference')
    if measure == 'RR':
        if scale == 'log-transform':
            ax.hlines(0, 0, np.max(r['timeline'] + 0.5), linewidth=1.5)
            ax.set_ylabel('ln(Risk Ratio)')
        elif scale == 'log':
            ax.set_ylabel('Risk Ratio')
            ax.set_yscale('log')
            ax.yaxis.set_major_formatter(mticker.ScalarFormatter())
            ax.yaxis.get_major_formatter().set_scientific(False)
            ax.yaxis.get_major_formatter().set_useOffset(False)
            ax.hlines(1, 0, np.max(r['timeline'] + 0.5), linewidth=1.5)
        else:
            ax.hlines(1, 0, np.max(r['timeline'] + 0.5), linewidth=1.5)
    ax.set_xlabel('Time')
    ax.set_xlim([0, np.max(r['timeline']) + 0.5])
    return ax
